/**
 * @file private.cl
 * Remeshing function, vectorized, private variable.
 */

void remesh(uint i, float dx, float invdx, float__N__ s, float__N__ p, __local float* gscal_loc);


/**
 * Remesh particles in local buffer.
 *
 * Remeshing formula is given a compiling time.
 * Use of builtin OpenCL functions fma and mix. Computations through OpenCL vector types.
 * Use of a private temporary variable for remeshing weights.
 *
 * @param i Particle index
 * @param dx Space step
 * @param invdx 1/dx
 * @param s Particle scalar
 * @param p Particle position
 * @param gscal_loc Local buffer for result
 *
 * @remark <code>NB_I</code>, <code>NB_II</code>, <code>NB_III</code> : points number in directions from 1st varying index to last.
 * @remark <code>__N__</code> is expanded at compilation time by vector width.
 * @remark <code>__NN__</code> is expanded at compilation time by a sequence of integer for each vector component.
 * @remark <code>FORMULA</code> : remeshing formula flag {<code>M4PRIME</code>, <code>M6PRIME</code>, <code>M8PRIME</code>, <code>L6STAR</code>}
 * @see parmepy.gpu.tools.parse_file
 * @see parmepy.gpu.cl_src.common
 */
void remesh(uint i, float dx, float invdx, float__N__ s, float__N__ p, __local float* gscal_loc){
  float__N__ y,			/* Normalized distance to nearest left grid point */
    temp;			/* Temporary remeshing weights */
  int__N__ ind;			/* Integer coordinate */
  uint__N__ index;		/* Remeshing index */

  ind = convert_int__N___rtn(p * invdx);
  y = (p - convert_float__N__(ind) * dx) * invdx;

#if FORMULA == M4PRIME
  index = convert_uint__N__((ind - 1 + NB_I) % NB_I);
#elif FORMULA == M6PRIME
  index = convert_uint__N__((ind - 2 + NB_I) % NB_I);
#elif FORMULA == M8PRIME || FORMULA == L6STAR
  index = convert_uint__N__((ind - 3 + NB_I) % NB_I);
#endif

  temp = alpha(y) * s;
  gscal_loc[noBC_id(index.s__NN__)] += temp.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  temp = beta(y) * s;
  gscal_loc[noBC_id(index.s__NN__)] += temp.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  temp = gamma(y) * s;
  gscal_loc[noBC_id(index.s__NN__)] += temp.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  temp = delta(y) * s;
  gscal_loc[noBC_id(index.s__NN__)] += temp.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

#if FORMULA == M6PRIME || FORMULA == M8PRIME
  index = (index + 1) % NB_I;
  temp = eta(y) * s;
  gscal_loc[noBC_id(index.s__NN__)] += temp.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  temp = zeta(y) * s;
  gscal_loc[noBC_id(index.s__NN__)] += temp.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

#if FORMULA == M8PRIME || FORMULA == L6STAR
  index = (index + 1) % NB_I;
  temp = theta(y) * s;
  gscal_loc[noBC_id(index.s__NN__)] += temp.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  temp = iota(y) * s;
  gscal_loc[noBC_id(index.s__NN__)] += temp.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

#endif
#endif
}
