/**
 * @file private_vector_3d.cl
 * Remeshing function, vectorized, private variable for 3D vector remeshing.
 */

void remesh(uint i, float dx, float invdx,
	    float__N__ v_X, float__N__ v_Y,float__N__ v_Z,
	    float__N__ p,
	    __local float* gvec_X_loc, __local float* gvec_Y_loc, __local float* gvec_Z_loc);


/**
 * Remesh particles in local buffer.
 *
 * Remeshing formula is given a compiling time.
 * Use of builtin OpenCL functions fma and mix. Computations through OpenCL vector types.
 * Use of a private temporary variable for remeshing weights.
 *
 * @param i Particle index
 * @param dx Space step
 * @param invdx 1/dx
 * @param s Particle scalar
 * @param p Particle position
 * @param gscal_loc Local buffer for result
 *
 * @remark <code>NB_I</code>, <code>NB_II</code>, <code>NB_III</code> : points number in directions from 1st varying index to last.
 * @remark <code>__N__</code> is expanded at compilation time by vector width.
 * @remark <code>__NN__</code> is expanded at compilation time by a sequence of integer for each vector component.
 * @remark <code>FORMULA</code> : remeshing formula flag {<code>M4PRIME</code>, <code>M6PRIME</code>, <code>M8PRIME</code>, <code>L6STAR</code>}
 * @see parmepy.gpu.tools.parse_file
 * @see parmepy.gpu.cl_src.common
 */
void remesh(uint i, float dx, float invdx,
	    float__N__ v_X, float__N__ v_Y,float__N__ v_Z,
	    float__N__ p,
	    __local float* gvec_X_loc, __local float* gvec_Y_loc, __local float* gvec_Z_loc){
  float__N__ y,			/* Normalized distance to nearest left grid point */
    w;			        /* Temporary remeshing weights */
  int__N__ ind;			/* Integer coordinate */
  uint__N__ index;		/* Remeshing index */

  ind = convert_int__N___rtn(p * invdx);
  y = (p - convert_float__N__(ind) * dx) * invdx;

#if FORMULA == M4PRIME
  index = convert_uint__N__((ind - 1 + NB_I) % NB_I);
#elif FORMULA == M6PRIME
  index = convert_uint__N__((ind - 2 + NB_I) % NB_I);
#elif FORMULA == M8PRIME || FORMULA == L6STAR
  index = convert_uint__N__((ind - 3 + NB_I) % NB_I);
#endif

  w = alpha(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  gvec_Z_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Z.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = beta(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  gvec_Z_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Z.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = gamma(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  gvec_Z_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Z.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = delta(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  gvec_Z_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Z.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

#if FORMULA == M6PRIME || FORMULA == M8PRIME
  index = (index + 1) % NB_I;
  w = eta(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  gvec_Z_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Z.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = zeta(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  gvec_Z_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Z.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

#if FORMULA == M8PRIME || FORMULA == L6STAR
  index = (index + 1) % NB_I;
  w = theta(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  gvec_Z_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Z.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

  index = (index + 1) % NB_I;
  w = iota(y);
  gvec_X_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_X.s__NN__;
  gvec_Y_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Y.s__NN__;
  gvec_Z_loc[noBC_id(index.s__NN__)] += w.s__NN__ * v_Z.s__NN__;
  barrier(CLK_LOCAL_MEM_FENCE);

#endif
#endif
}
