"""
@file gpu_analytic.py
Analytic operator on GPU.
"""
from parmepy.operator.discrete.discrete import DiscreteOperator
from parmepy.constants import debug, np
from parmepy import __VERBOSE__
from parmepy.gpu.tools import get_opencl_environment
from parmepy.gpu.gpu_kernel import KernelLauncher
#from parmepy.gpu import CL_PROFILE
from parmepy.tools.timers import timed_function


class GPUAnalytic(DiscreteOperator):
    """
    Analytic operator on GPU.

    Evaluates a formula on a GPU field.
    """

    @debug
    def __init__(self, variables,
                 platform_id=0, device_id=0,
                 device_type='gpu',
                 method='', src=None, precision=None):
        """
        Constructor.

        Computes the OpenCL environment with
        parmepy.gpu.tools.get_opencl_environment.

        @param variables a list of discrete variables
        @param platform_id OpenCL platform id
        @param device_id OpenCL device id
        @param device_type OpenCL device type
        @param method Method to use
        @param src User OpenCL source file
        @param precision Floating point precision
        """
        DiscreteOperator.__init__(self, variables, method, name="Analytic_GPU")
        ## OpenCL work-item number
        self.workItemNumber = 0
        ## Operated field
        self.field = self.variables[0]
        ## User source file
        self.usr_src = src
        ## Field resolution
        self.resolution = self.field.topology.mesh.resolution
        ## OpenCL environment
        self.cl_env = get_opencl_environment(platform_id, device_id,
                                             device_type, precision,
                                             self.resolution)

    @debug
    def setUp(self):
        """
        Set up.
        Compute OpenCL work-item number and space index.\n
        Compile OpenCL sources. The OpenCL kernel mus be named as follows :
        <code>analytic"FieldName"</code> with <code>"FieldName"</code> is the
        name of the field as given by user.
        """
        ## Floating point precision
        self.gpu_precision = self.cl_env.precision
        ## OpenCL space index
        self.workItemNumber, self.gwi, self.lwi = self.cl_env.get_WorkItems(
            self.resolution)
        if __VERBOSE__:
            print "Work-items basic config : ",
            print self.workItemNumber, self.gwi, self.lwi
        dim = self.field.topology.dim
        ## Minimum field coordinate (as GPU float4)
        self.coord_min = np.ones(4, dtype=self.gpu_precision)
        ## Field mesh size (as GPU float4)
        self.mesh_size = np.ones(4, dtype=self.gpu_precision)
        self.coord_min[0:dim] = np.asarray(self.field.topology.mesh.origin,
                                           dtype=self.gpu_precision)
        self.mesh_size[0:dim] = np.asarray(self.field.topology.mesh.space_step,
                                           dtype=self.gpu_precision)
        # Kernel sources
        if __VERBOSE__:
            print "=== Kernel sources ==="
        ## OpenCL compiling options
        self.build_options = ""
        self.build_options += " -D WI_NB=" + str(self.workItemNumber)
        ## OpenCL Binaries
        self.prg = self.cl_env.build_src(self.usr_src, self.build_options)
        kernel_name = 'analytic' + self.field.name.split('_D')[0]
        self.numMethod = KernelLauncher(eval('self.prg.' + kernel_name),
                                        self.cl_env.queue,
                                        self.gwi,
                                        self.lwi)

    @debug
    @timed_function
    def apply(self, simulation):
        t = self.gpu_precision(simulation.time)
        self.field.initialize(self.numMethod, t)
        for df in self.output:
            df.data_on_device = True
            df.contains_data = False

    def finalize(self):
        if self.numMethod.f_timer is not None:
            for f_timer in self.numMethod.f_timer:
                self.timer.addFunctionTimer(f_timer)
        DiscreteOperator.finalize(self)
