"""
@file gpu_discrete.py

Contains class for discrete fields on GPU.
"""
from abc import ABCMeta, abstractmethod

from parmepy import __VERBOSE__
from parmepy.constants import ORDER, np, debug
from parmepy.fields.scalar import ScalarField
from parmepy.fields.vector import VectorField
from parmepy.gpu import cl, PARMES_REAL_GPU, CL_PROFILE
from parmepy.tools.timers import timed_function
from parmepy.gpu.gpu_kernel import KernelLauncher, KernelListLauncher


class GPUDiscreteField(object):
    """
    GPU discrete field interface.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, queue, precision):
        """
        Constructor.
        @param queue : OpenCL queue
        @param precision : Required numerical precision.
        """
        ## OpenCL queue
        self.queue = queue
        ## OpenCL Buffer pointer
        self.gpu_data = None
        ## Precision for the field
        self.precision = precision
        ## Memory used
        self.mem_size = 0
        ## Flag for updated data localization
        self.data_on_device = False
        ## Initialization OpenCL kernel as KernelLauncher
        self.init_kernel = None
        self._isReleased = False

    def setInitializationKernel(self, kernel):
        """
        Set the initialization kernel
        @param kernel : KernelLauncher to use for initialize field.
        """
        self.init_kernel = kernel

    @abstractmethod
    def toHost(self):
        pass

    @abstractmethod
    def toDevice(self):
        pass


class GPUVectorField(VectorField, GPUDiscreteField):
    """
    GPU Discrete vector field implementation.
    Allocates OpenCL device memory for the field.
    """
    def __init__(self, queue, parent,
                 topology=None, name="?", idFromParent=None,
                 precision=PARMES_REAL_GPU, layout=True):
        """
        Constructor.
        @param queue : OpenCL queue
        @param precision : Floating point precision
        @param parent : Continuous field.
        @param topology : Topology informations
        @param name : Field name
        @param idFromParent : Index in the parent's discrete fields
        @param layout : Boolean indicating if components are arranged in memory
        @see parmepy.fields.vector.VectorField.__init__
        """
        super(GPUVectorField, self).__init__(self, queue, precision,
                                             parent, topology,
                                             name, idFromParent)
        self.gpu_data = [None for d in xrange(self.dimension)]
        self.layout = layout
        for d in xrange(self.dimension):
            self.data[d] = np.asarray(self.data[d],
                                      dtype=precision, order=ORDER)
            self.gpu_data[d] = cl.Buffer(self.queue.context,
                                         cl.mem_flags.READ_WRITE,
                                         size=self.data[d].nbytes)
            self.mem_size += self.gpu_data[d].size
        if __VERBOSE__:
            print self.name, self.mem_size, "Bytes (",
            print self.mem_size / (1024 ** 2), "MB)"

    @classmethod
    def fromVectorField(cls, queue, vfield, precision=PARMES_REAL_GPU,
                        layout=True):
        """
        Contructor from a discrete vector field.
        Mutates the given VectorField to a GPUVectorField.
        @param cls : Class of the class method (GPUVectorField)
        @param queue : OpenCL queue
        @param vfield : VectorField
        @param precision : Floating point precision
        @param layout : Boolean indicating if components are arranged in memory
        """
        if not isinstance(vfield, GPUDiscreteField):
            vfield.__class__ = cls
            GPUDiscreteField.__init__(vfield, queue, precision)
            vfield.gpu_data = [None for d in xrange(vfield.dimension)]
            vfield.layout = layout
            for d in xrange(vfield.dimension):
                vfield.data[d] = np.asarray(vfield.data[d],
                                            dtype=precision, order=ORDER)
                vfield.gpu_data[d] = cl.Buffer(vfield.queue.context,
                                               cl.mem_flags.READ_WRITE,
                                               size=vfield.data[d].nbytes)
                vfield.mem_size += vfield.gpu_data[d].size
            if __VERBOSE__:
                print vfield.name, vfield.mem_size, "Bytes (",
                print vfield.mem_size / (1024 ** 2), "MB)"

    @debug
    @timed_function
    def dump(self, filename, mode=None):
        self.toHost()
        VectorField.dump(self, filename, mode)

    @debug
    @timed_function
    def load(self, filename):
        VectorField.load(self, filename)
        self.toDevice()
        self.contains_data = False
        self.data_on_device = True

    @debug
    @timed_function
    def initialize(self, formula=None, *args):
        """
        GPU data initialization.
        Performs the initialization from different ways if device not already
        contains up-to-date data:
          - with an OpenCL kernel,
          - with a python formula (as VectorField) and the copy data to device.
        @param formula : Formula to use.
        @param args : formula extra parameters
        """
        isGPUKernel = isinstance(formula, KernelLauncher) \
            or isinstance(formula, KernelListLauncher)
        if not isGPUKernel and self.init_kernel is None:
            VectorField.initialize(self, formula, *args)
            if self.contains_data:
                if self.dimension == 2:
                    self.data[0] = np.asarray(
                        self.data[0],
                        dtype=self.precision, order=ORDER)
                    self.data[1] = np.asarray(
                        self.data[1],
                        dtype=self.precision, order=ORDER)
                else:
                    self.data[0] = np.asarray(
                        self.data[0],
                        dtype=self.precision, order=ORDER)
                    self.data[1] = np.asarray(
                        self.data[1],
                        dtype=self.precision, order=ORDER)
                    self.data[2] = np.asarray(
                        self.data[2],
                        dtype=self.precision, order=ORDER)
                self.toDevice()
        else:
            if isGPUKernel:
                self.init_kernel = formula
            print "Using OpenCL kernel", self.init_kernel
            coord_min = np.ones(4, dtype=self.precision)
            mesh_size = np.ones(4, dtype=self.precision)
            coord_min[0:self.dimension] = np.asarray(
                self.topology.mesh.origin,
                dtype=self.precision)
            mesh_size[0:self.dimension] = np.asarray(
                self.topology.mesh.space_step,
                dtype=self.precision)
            if self.dimension == 2:
                self.init_kernel(self.gpu_data[0],
                                 self.gpu_data[1],
                                 coord_min, mesh_size, *args)
            else:
                self.init_kernel(self.gpu_data[0],
                                 self.gpu_data[1],
                                 self.gpu_data[2],
                                 coord_min, mesh_size, *args)
        self.contains_data = False
        self.data_on_device = True

    def release(self):
        if not self._isReleased:
            if __VERBOSE__:
                print "deallocate :", self.name,
                print " (" + str(self.mem_size / (1024. ** 2)) + " MBytes)"
            [self.gpu_data[d].release()
             for d in xrange(len(self.resolution))]
            if self.init_kernel is not None:
                for f_timer in self.init_kernel.f_timer:
                    self.timer.addFunctionTimer(f_timer)
            print self.timer
            self._isReleased = True

    @timed_function
    def toDevice(self):
        """
        Host to device method.

        Performs a direct OpenCL copy from numpy arrays
        to OpenCL Buffers.\n
        Arrange memory on device so that vector components are
        contiguous in the direction of the component, if layout flag is True.\n
        Example : A 3D vector field F(x,y,z) is made up of 3
        OpenCL Buffers Fx, Fy, Fz. The memory layout is :
        - Fx : x-major ordering. On device,
        Fx[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fx(i,j,k)
        - Fy : y-major ordering. On device,
        Fy[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fy(j,i,k)
        - Fz : z-major ordering. On device,
        Fz[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fz(k,i,j)
        """
        if __VERBOSE__:
            print "host->device :", self.name,
        time = 0.
        evt = [None, None, None]
        if self.layout:
            evt[0] = cl.enqueue_copy(
                self.queue, self.gpu_data[0],
                self.data[0].ravel(order=ORDER))
            evt[1] = cl.enqueue_copy(
                self.queue, self.gpu_data[1],
                self.data[1].swapaxes(0, 1).ravel(order=ORDER))
            if self.dimension == 3:
                evt[2] = cl.enqueue_copy(
                    self.queue, self.gpu_data[2],
                    self.data[2].swapaxes(0, 1).swapaxes(0, 2).ravel(
                        order=ORDER))
        else:
            evt[0] = cl.enqueue_copy(
                self.queue, self.gpu_data[0],
                self.data[0].ravel(order=ORDER))
            evt[1] = cl.enqueue_copy(
                self.queue, self.gpu_data[1],
                self.data[1].ravel(order=ORDER))
            if self.dimension == 3:
                evt[2] = cl.enqueue_copy(
                    self.queue, self.gpu_data[2],
                    self.data[2].ravel(order=ORDER))
        if CL_PROFILE:
            self.queue.finish()
            for e in evt:
                if e is not None:
                    time += (e.profile.end - e.profile.start) * 1e-9
        if __VERBOSE__ and CL_PROFILE:
            print self.mem_size, "Bytes transfered at ",
            print "{0:.3f} GBytes/sec".format(
                self.mem_size / (time * 1024 ** 3))
            print "Transfer host->device"

    @timed_function
    def toHost(self):
        """
        Device to host method.

        Performs a direct OpenCL copy from OpenCL Buffers
        to numpy arrays.\n
        As memory layout, if set, is arranged on device, not only a
        copy is performed but also transpositions to have numpy
        arrays consistent to each other.
        """
        if not self.contains_data:
            if __VERBOSE__:
                print "device->host :", self.name,
            self.queue.finish()
            shape = self.data[0].shape
            time = 0.
            evt = [None, None, None]
            if self.layout:
                if self.dimension == 3:
                    # Get ZDIR component
                    temp = np.empty_like(self.data[2]).resize(
                        (shape[2], shape[0], shape[1]))
                    evt[2] = cl.enqueue_copy(self.queue, temp,
                                             self.gpu_data[2])
                    self.data[2] = temp.swapaxes(0, 2).swapaxes(0, 1)
                    # Get YDIR component
                    temp = np.empty_like(self.data[1]).resize(
                        (shape[1], shape[0], shape[2]))
                    evt[1] = cl.enqueue_copy(self.queue, temp,
                                             self.gpu_data[1])
                    self.data[1] = temp.swapaxes(0, 1)
                    # Get XDIR component
                    evt[0] = cl.enqueue_copy(self.queue, self.data[0],
                                             self.gpu_data[0])
                else:
                    # Get YDIR component
                    temp = np.empty_like(self.data[1]).resize(
                        (shape[1], shape[0]))
                    print "ICICICICIIC ", temp.shape
                    evt[1] = cl.enqueue_copy(self.queue, temp,
                                             self.gpu_data[1])
                    self.data[1] = temp.swapaxes(0, 1)
                    # Get XDIR component
                    evt[0] = cl.enqueue_copy(self.queue, self.data[0],
                                             self.gpu_data[0])
            else:
                if self.dimension == 3:
                    # Get ZDIR component
                    evt[2] = cl.enqueue_copy(self.queue, self.data[2],
                                             self.gpu_data[2])
                    # Get YDIR component
                    evt[1] = cl.enqueue_copy(self.queue, self.data[1],
                                             self.gpu_data[1])
                    # Get XDIR component
                    evt[0] = cl.enqueue_copy(self.queue, self.data[0],
                                             self.gpu_data[0])
                else:
                    # Get YDIR component
                    evt[1] = cl.enqueue_copy(self.queue, self.data[1],
                                             self.gpu_data[1])
                    # Get XDIR component
                    evt[0] = cl.enqueue_copy(self.queue, self.data[0],
                                             self.gpu_data[0])
            if CL_PROFILE:
                self.queue.finish()
                for e in evt:
                    if e is not None:
                        time += (e.profile.end - e.profile.start) * 1e-9
            if __VERBOSE__ and CL_PROFILE:
                print self.mem_size, "Bytes transfered at ",
                print "{0:.3f} GBytes/sec".format(
                    self.mem_size / (time * 1024 ** 3))
                print 'Transfer device->host'


class GPUScalarField(ScalarField, GPUDiscreteField):
    """
    GPU Discrete scalar field implementation.
    """
    def __init__(self, queue, parent,
                 topology=None, name="?", idFromParent=None,
                 precision=PARMES_REAL_GPU):
        """
        Constructor.
        @param queue : OpenCL queue
        @param precision : Floating point precision
        @param parent : Continuous field.
        @param topology : Topology informations
        @param name : Field name
        @param idFromParent : Index in the parent's discrete fields
        @see parmepy.fields.vector.ScalarField.__init__
        """
        super(GPUScalarField, self).__init__(self, queue, precision,
                                             parent, topology,
                                             name, idFromParent)
        self.data = np.asarray(self.data,
                               dtype=precision, order=ORDER)
        self.gpu_data = cl.Buffer(self.queue.context,
                                  cl.mem_flags.READ_WRITE,
                                  size=self.data.nbytes)
        self.mem_size = self.gpu_data.size
        if __VERBOSE__:
            print self.name, self.mem_size, "Bytes (",
            self.mem_size / (1024 ** 2), "MB)"

    @classmethod
    def fromScalarField(cls, queue, sfield, precision=PARMES_REAL_GPU):
        """
        Contructor from a discrete scalar field.

        Mutates the given ScalarField to a GPUScalarField.

        @param cls : Class
        @param queue : OpenCL queue
        @param sfield : ScalarField
        @param precision : Floating point precision
        """
        if not isinstance(sfield, GPUDiscreteField):
            sfield.__class__ = cls
            GPUDiscreteField.__init__(sfield, queue, precision)
            sfield.data = np.asarray(sfield.data,
                                     dtype=precision, order=ORDER)
            sfield.gpu_data = cl.Buffer(sfield.queue.context,
                                        cl.mem_flags.READ_WRITE,
                                        size=sfield.data.nbytes)
            sfield.mem_size = sfield.gpu_data.size
            if __VERBOSE__:
                print sfield.name, sfield.mem_size, "Bytes (",
                print sfield.mem_size / (1024 ** 2), "MB)"

    @debug
    def dump(self, filename, mode=None):
        self.toHost()
        ScalarField.dump(self, filename, mode)

    @debug
    def load(self, filename):
        ScalarField.load(self, filename)
        self.toDevice()
        self.contains_data = False
        self.data_on_device = True

    @debug
    @timed_function
    def initialize(self, formula=None, *args):
        """
        GPU data initialization.
        @param formula : Python formula to use
        @param args : Extra parameters
        Performs the initialization from different ways if device not already
        contains up-to-date data:
          - with an OpenCL kernel,
          - with a python formula (as VectorField) and the copy data to device.
        """
        isGPUKernel = isinstance(formula, KernelLauncher) \
            or isinstance(formula, KernelListLauncher)
        if not isGPUKernel and self.init_kernel is None:
            ScalarField.initialize(self, formula, *args)
            if self.contains_data:
                self.data = np.asarray(self.data,
                                       dtype=self.precision, order=ORDER)
                self.toDevice()
        else:
            if isGPUKernel:
                self.init_kernel = formula
            coord_min = np.ones(4, dtype=self.precision)
            mesh_size = np.ones(4, dtype=self.precision)
            coord_min[0:self.dimension] = np.asarray(
                self.topology.mesh.origin,
                dtype=self.precision)
            mesh_size[0:self.dimension] = np.asarray(
                self.topology.mesh.space_step,
                dtype=self.precision)
            self.init_kernel(self.gpu_data, coord_min, mesh_size, *args)
        self.contains_data = False
        self.data_on_device = True

    def release(self):
        if not self._isReleased:
            if __VERBOSE__:
                print "deallocate :", self.name,
                print " (" + str(self.mem_size / (1024. ** 2)) + " MBytes)"
                self.gpu_data.release()
            if self.init_kernel is not None:
                for f_timer in self.init_kernel.f_timer:
                    self.timer.addFunctionTimer(f_timer)
            print self.timer
            self._isReleased = True

    @timed_function
    def toDevice(self):
        """
        Host to device method.

        Performs a direct OpenCL copy from numpy arrays
        to OpenCL Buffers.\n
        Arrange memory on device so that vector components are
        contiguous in the direction of the component.\n
        Example : A 3D vector field F(x,y,z) is made up of 3
        OpenCL Buffers Fx, Fy, Fz. The memory layout is :
        - Fx : x-major ordering. On device,
        Fx[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fx(i,j,k)
        - Fy : y-major ordering. On device,
        Fy[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fy(j,i,k)
        - Fz : z-major ordering. On device,
        Fz[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fz(k,i,j)
        """
        if __VERBOSE__:
            print "host->device :", self.name,
        time = 0.
        evt = [None, None, None]
        evt[0] = cl.enqueue_copy(self.queue, self.gpu_data,
                                 self.data.ravel(order=ORDER))
        if CL_PROFILE:
            self.queue.finish()
            for e in evt:
                if e is not None:
                    time += (e.profile.end - e.profile.start) * 1e-9
        if __VERBOSE__ and CL_PROFILE:
            print self.mem_size, "Bytes transfered at ",
            print "{0:.3f} GBytes/sec".format(
                self.mem_size / (time * 1024 ** 3))

    @timed_function
    def toHost(self):
        """
        Device to host method.

        Performs a direct OpenCL copy from OpenCL Buffers
        to numpy arrays.\n
        As memory layout is arranged on device, not only a
        copy is performed but also transpositions to have numpy
        arrays consistent to each other.
        """
        if not self.contains_data:
            if __VERBOSE__:
                print "device->host :", self.name,
            time = 0.
            evt = [None, None, None]
            evt[0] = cl.enqueue_copy(self.queue, self.data,
                                     self.gpu_data)
            if CL_PROFILE:
                self.queue.finish()
                for e in evt:
                    if e is not None:
                        time += (e.profile.end - e.profile.start) * 1e-9
            if __VERBOSE__ and CL_PROFILE:
                print self.mem_size, "Bytes transfered at ",
                print "{0:.3f} GBytes/sec".format(
                    self.mem_size / (time * 1024 ** 3))
