"""
@file gpu_particle_advection.py

Discrete advection representation
"""
from abc import ABCMeta, abstractmethod
from parmepy import __VERBOSE__
from parmepy.constants import np, debug
from parmepy.operator.discrete.particle_advection import ParticleAdvection
from parmepy.gpu import PARMES_REAL_GPU, cl
from parmepy.gpu.tools import get_opencl_environment
from parmepy.gpu.gpu_kernel import KernelLauncher
#from parmepy.gpu.gpu_kernel import KernelListLauncher
#from parmepy.numerics.splitting import Splitting


class GPUParticleAdvection(ParticleAdvection):
    """
    Particle advection operator representation on GPU.

    """
    __metaclass__ = ABCMeta

    @debug
    @abstractmethod
    def __init__(self, velocity, advectedFields, d,
                 part_position=None, part_advectedFields=None,
                 platform_id=0, device_id=0,
                 device_type='gpu',
                 method='', src=None, precision=PARMES_REAL_GPU):
        """
        Create a Advection operator.
        Work on a given field (scalar or vector) at a given velocity to compute
        advected values.
        OpenCL kernels are build once per dimension in order to handle
        directional splitting with resolution non uniform in directions.

        @param velocity : Velocity field
        @param advectedFields : Advected fields
        @param d : Direction to advect
        @param platform_id : OpenCL platform id (default = 0).
        @param device_id : OpenCL device id (default = 0).
        @param device_type : OpenCL device type (default = 'gpu').
        @param method : the method to use. {'m4prime', 'm6prime', 'm8prime',
        'l6star'}
        @param src : User OpenCL sources.
        @param precision : Floating point precision required (default = Single
        Precision)
        @param splittingConfig : Directional splitting configuration
        (parmepy.numerics.splitting.Splitting.__init__)
        """
        ParticleAdvection.__init__(self, velocity, advectedFields, d,
                                   part_position, part_advectedFields, method)
        if method.find('m4prime') >= 0:
            self.method = "m4prime"
        elif method.find('m6prime') >= 0:
            self.method = "m6prime"
        elif method.find('m8prime') >= 0:
            self.method = "m8prime"
        elif method.find('l6star') >= 0:
            self.method = "l6star"
        elif method.find('l4star') >= 0:
            self.method = "l4star"
        else:
            print 'No remeshing fromula specified, use default m6prime'
            self.method = "m6prime"
        self.name += ' GPU ' + self.method
        self.user_gpu_src = src
        self.num_method = None

        # Compute resolutions for kernels for each direction.
        resol = self.advectedFields[0].topology.mesh.resolution
        self.resolution = resol.copy()
        # resolution[0] = NB_X, NB_Y (, NB_Z) for X direction
        # resolution[1] = NB_Y, NB_X (, NB_Z) for Y direction
        # resolution[2] = NB_Z, NB_X, NB_Y for Z direction
        if len(resol) == 2:
            if self.dir == 1:
                self.resolution[0] = resol[1]
                self.resolution[1] = resol[0]
        else:
            if self.dir == 1:
                self.resolution[0] = resol[1]
                self.resolution[1] = resol[0]
                self.resolution[2] = resol[2]
            if self.dir == 2:
                self.resolution[0] = resol[2]
                self.resolution[1] = resol[0]
                self.resolution[2] = resol[1]

        self.cl_env = get_opencl_environment(platform_id, device_id,
                                             device_type, precision,
                                             resol)
        self.warning = True
        self.prg = None

    @debug
    def setUp(self):
        """
        Set up.
        Compute OpenCL work-item number and space index.\n
        Compile OpenCL sources.
        Allocates OpenCL buffers. Buffers are initialized with OpenCL kernels
        if named as follows : <code>init"FieldName"</code> with
        <code>"FieldName"</code> the field
        name as given by user.
        """
        self.gpu_precision = self.cl_env.precision
        self.workItemNumber, self.gwi, self.lwi = self.cl_env.get_WorkItems(
            self.resolution)
        if __VERBOSE__:
            print "Work-items basic config : ",
            print self.workItemNumber, self.gwi, self.lwi
        dim = self.advectedFields[0].topology.dim
        self.coord_min = np.ones(4, dtype=self.gpu_precision)
        self.mesh_size = np.ones(4, dtype=self.gpu_precision)
        self.coord_min[0:dim] = np.asarray(
            self.advectedFields[0].topology.mesh.origin,
            dtype=self.gpu_precision)
        self.mesh_size[0:dim] = np.asarray(
            self.advectedFields[0].topology.mesh.space_step,
            dtype=self.gpu_precision)
        self.build_options = ""
        self._collect_usr_cl_src(self.user_gpu_src)
        self._collect_kernels_cl_src_copy()
        self._collect_kernels_cl_src_transpositions()
        self._collect_kernels_cl_src()
        if __VERBOSE__:
            print "=== OpenCL Buffer allocations ==="
        self._buffer_allocations()
        if __VERBOSE__:
            print "===\n"
            print "=== OpenCL Buffer initialisation ==="
        if self.dir == 0:
            self._buffer_initialisations()
        if __VERBOSE__:
            print "===\n"
        if self.advectedFields[0].isVector:
            if dim == 2:
                self.apply = self._apply_in_dir_vector_2d
                #self.numMethod = Splitting(self._apply_in_dir_vector_2d,
                #                           self.advectedFields[0].dimension,
                #                           config=self.splittingConfig)
            else:
                self.apply = self._apply_in_dir_vector_3d
                #self.numMethod = Splitting(self._apply_in_dir_vector_3d,
                #                           self.advectedFields[0].dimension,
                #                           config=self.splittingConfig)
        else:
            self.apply = self._apply_in_dir
            #self.numMethod = Splitting(self._apply_in_dir,
            #                           self.advectedFields[0].dimension,
            #                           config=self.splittingConfig)

    def _buffer_initialisations(self):
        """
        OpenCL buffer initializations from user OpenCL kernels.
        Looking for kernels named <code>init<FieldName></code>.
        """
        for gpudf in self.variables:
            match = 'init' + gpudf.name
            # Looking for initKernel
            if not self.prg is None:
                for k in self.prg.all_kernels():
                    k_name = k.get_info(cl.kernel_info.FUNCTION_NAME)
                    if match.find(k_name) >= 0:
                        if __VERBOSE__:
                            print gpudf.name, '-> OpenCL Kernel', k_name
                        gpudf.setInitializationKernel(KernelLauncher(
                            cl.Kernel(self.prg, k_name), self.cl_env.queue,
                            self.gwi, self.lwi))

    def _collect_kernels_cl_src_copy(self):
        """
        Compile OpenCL sources for copy kernel.

        Kernel configuration :
        <table border>
        <tr>
           <td><b>Dimension</b></td>
           <td><b>Precision</b></td>
           <td><b>kernel sources</b></td>
           <td><b>Global space index</b></td>
           <td><b>Local space index</b></td>
           <td><b>Tile</b></td>
           <td><b>Block rows</b></td>
        </tr>
        <tr>
           <td rowspan="2">3</td>
           <td>Single</td>
           <td>copy_3D_opt4.cl parmepy.gpu.cl_src.copy_3D_opt4.cl</td>
           <td>(Nx/4, Ny/2, Nz)</td>
           <td>(4, 8, 1)</td>
           <td>16**2</td>
           <td>8</td>
        </tr>
        <tr>
           <td>Double</td>
           <td>copy_3D_locMem.cl</td>
           <td>(Nx, Ny/4, Nz)</td>
           <td>(32, 8, 1)</td>
           <td>32**2</td>
           <td>8</td>
        </tr>
        <tr>
           <td rowspan="2">2</td>
           <td>Single</td>
           <td>copy_2D_opt2.cl</td>
           <td>(Nx/2, Ny/2)</td>
           <td>(8, 8)</td>
           <td>16**2</td>
           <td>8</td>
        </tr>
        <tr>
           <td>Double</td>
           <td>copy_2D_opt2.cl</td>
           <td>(Nx/2, Ny/16)</td>
           <td>(16, 2)</td>
           <td>32**2</td>
           <td>2</td>
        </tr>
        </table>
        """
        build_options = self.build_options
        # copy settings
        if len(self.resolution) == 3:
            if self.gpu_precision is PARMES_REAL_GPU:
                vec = 4
                src_copy = 'kernels/copy.cl'
                t_dim = 16
                b_rows = 8
            else:
                vec = 1
                src_copy = 'kernels/copy_locMem.cl'
                t_dim = 32
                b_rows = 8
        else:
            vec = 2
            if self.gpu_precision is PARMES_REAL_GPU:
                src_copy = 'kernels/copy.cl'
                t_dim = 16
                b_rows = 8
            else:
                src_copy = 'kernels/copy.cl'
                t_dim = 32
                b_rows = 2
        # Build code
        build_options += " -D TILE_DIM_COPY={0}".format(t_dim)
        build_options += " -D BLOCK_ROWS_COPY={0}".format(b_rows)
        # Direction specific settings
        # prg = len(self.resolution) * [None]
        # gwi = len(self.resolution) * [None]
        # lwi = len(self.resolution) * [None]
        gwi = (int(self.resolution[0] / vec),
               int(b_rows * self.resolution[1] / t_dim),
               int(1 if len(self.resolution) == 2 else self.resolution[2]))
        lwi = (t_dim / vec, b_rows, 1)
        prg = self.cl_env.build_src(
            src_copy,
            build_options + self.cl_env.default_sizes_constants[self.dir],
            vec)
        self.init_copy = KernelLauncher(prg.copy,
                                        self.cl_env.queue, gwi, lwi)

    def _collect_kernels_cl_src_transpositions(self):
        """Compile OpenCL sources for transpositions kernel.
        @remark : Transpositions kernels are launched as initialization.
        Arrays are taken in their destination layout (for initialize in Y
        directions, either we came from X or Z but shapes are the Y ones).

        Kernel configuration :
        <table border>
        <tr>
           <td><b>Dimension</b></td>
           <td><b>Directions</b></td>
           <td><b>Precision</b></td>
           <td><b>kernel sources</b></td>
           <td><b>Global space index</b></td>
           <td><b>Local space index</b></td>
           <td><b>Tile</b></td>
           <td><b>Block rows</b></td>
        </tr>
        <tr>
           <td rowspan="4">3</td>
           <td rowspan="2">XY</td>
           <td>Single</td>
           <td>transpose_3D_xy_coalesced_locPad_Diag_2Dslice_opt2.cl</td>
           <td>(Nx/2, Ny/2, Nz)</td>
           <td>(8, 8, 1)</td>
           <td>16**2</td>
           <td>8</td>
        </tr>
        <tr>
           <td>Double</td>
           <td>transpose_3D_xy_coalesced_locPad_Diag_2Dslice_opt4.cl</td>
           <td>(Nx/4, Ny/8, Nz)</td>
           <td>(8, 4, 1)</td>
           <td>32**2</td>
           <td>4</td>
        </tr>
        <tr>
           <td rowspan="2">XZ</td>
           <td>Single</td>
           <td>transpose_3D_xz_coalesced_locPad_Diag_bis_3DBlock.cl</td>
           <td>(Nx, Ny/4, Nz/4)</td>
           <td>(16, 4, 4)</td>
           <td>16**3</td>
           <td>4</td>
        </tr>
        <tr>
           <td>Double</td>
           <td>transpose_3D_xz_coalesced_Diag_bis_3DBlock.cl</td>
           <td>(Nx, Ny/4, Nz/4)</td>
           <td>(8, 2, 2)</td>
           <td>8**3</td>
           <td>2</td>
        </tr>
        <tr>
           <td rowspan="2">2</td>
           <td rowspan="2">XY</td>
           <td>Single</td>
           <td>transpose_2D_xy_coalesced_locPad_Diag_opt4.cl</td>
           <td>(Nx/4, Ny/4)</td>
           <td>(8, 8)</td>
           <td>32**2</td>
           <td>8</td>
        </tr>
        <tr>
           <td>Double</td>
           <td>transpose_2D_xy_coalesced_locPad_Diag_opt4.cl</td>
           <td>(Nx/4, Ny/16)</td>
           <td>(8, 2)</td>
           <td>32**2</td>
           <td>2</td>
        </tr>
        </table>
        """
        # XY transposition settings
        build_options = self.build_options
        src_transpose_xy = 'kernels/transpose_xy.cl'
        build_options += " -D PADDING_XY=1"
        if len(self.resolution) == 3:
            if self.gpu_precision is PARMES_REAL_GPU:
                vec = 2
                t_dim = 16
                b_rows = 8
            else:
                vec = 4
                t_dim = 32
                b_rows = 4
        else:
            vec = 4
            if self.gpu_precision is PARMES_REAL_GPU:
                t_dim = 32
                b_rows = 8
            else:
                t_dim = 32
                b_rows = 2
        build_options += " -D TILE_DIM_XY={0}".format(t_dim)
        build_options += " -D BLOCK_ROWS_XY={0}".format(b_rows)
        # Settings are taken from destination layout as current layout.
        # gwi is computed form input layout (appears as transposed layout)
        gwi = (int(self.resolution[1] / vec),
               int(b_rows * self.resolution[0] / t_dim),
               int(1 if len(self.resolution) == 2 else self.resolution[2]))
        lwi = (t_dim / vec, b_rows, 1)
        prg = self.cl_env.build_src(
            src_transpose_xy,
            build_options + self.cl_env.default_sizes_constants[self.dir], vec)

        self.init_transpose_xy = KernelLauncher(
            prg.transpose_xy, self.cl_env.queue, gwi, lwi)

        if len(self.resolution) == 3:
            # XY transposition settings
            build_options = self.build_options
            src_transpose_xz = 'kernels/transpose_xz.cl'
            vec = 1
            if self.gpu_precision is PARMES_REAL_GPU:
                build_options += " -D PADDING_XZ=1"
                t_dim = 16
                b_rows = 4
            else:
                build_options += " -D PADDING_XZ=0"
                t_dim = 8
                b_rows = 2
            build_options += " -D TILE_DIM_XZ={0}".format(t_dim)
            build_options += " -D BLOCK_ROWS_XZ={0}".format(b_rows)
            # Settings are taken from destination layout as current layout.
            # gwi is computed form input layout (appears as transposed layout)
            gwi = (int(self.resolution[2] / vec),
                   int(b_rows * self.resolution[1] / t_dim),
                   int(b_rows * self.resolution[0] / t_dim))
            lwi = (t_dim / vec, b_rows, b_rows)
            prg = self.cl_env.build_src(
                src_transpose_xz,
                build_options + self.cl_env.default_sizes_constants[self.dir],
                vec)
            self.init_transpose_xz = KernelLauncher(
                prg.transpose_xz, self.cl_env.queue, gwi, lwi)

    def _collect_usr_cl_src(self, usr_src):
        """
        Build user sources.

        """
        build_options = self.build_options
        build_options += " -D WI_NB=" + str(self.workItemNumber)
        if usr_src is not None:
            self.prg = self.cl_env.build_src(usr_src, build_options)

    @debug
    def _apply_in_dir(self, simulation, dtCoeff, split_id, old_dir):
        """
        Apply operator along specified splitting direction.
        @param t : Current time
        @param dt : Time step
        @param d : Splitting direction
        @param split_id : Splitting step id
        """
        # Check new data on host
        for df in self.input:
            if df.contains_data and not df.data_on_device:
                df.toDevice()
        dim = self.advectedFields[0].dimension
        self.cl_env.queue.finish()
        # Particle init
        if old_dir == self.dir:
            self.init_copy(self.advectedFields[0].gpu_data,
                           self.part_advectedFields[0].gpu_data)
        else:
            if dim == 2:
                self.init_transpose_xy(self.advectedFields[0].gpu_data,
                                       self.part_advectedFields[0].gpu_data)
            else:
                if min(old_dir, self.dir) == 0:
                    self.init_transpose_xy(
                        self.advectedFields[0].gpu_data,
                        self.part_advectedFields[0].gpu_data)
                else:
                    self.init_transpose_xz(
                        self.advectedFields[0].gpu_data,
                        self.part_advectedFields[0].gpu_data)

    @debug
    def _apply_in_dir_vector_2d(self, simulation, dtCoeff, split_id, old_dir):
        """
        Apply operator along specified splitting direction. For 2D vector
        advection.
        @param t : Current time
        @param dt : Time step
        @param d : Splitting direction
        @param split_id : Splitting step id
        """
        # Check new data on host
        for df in self.input:
            if df.contains_data and not df.data_on_device:
                df.toDevice()
        self.cl_env.queue.finish()
        # Particle init
        if old_dir == self.dir:
            self.init_copy(self.advectedFields[0].gpu_data[0],
                           self.part_advectedFields[0].gpu_data[0])
            self.init_copy(self.advectedFields[0].gpu_data[1],
                           self.part_advectedFields[0].gpu_data[1])
        else:
            self.init_transpose_xy(
                self.advectedFields[0].gpu_data[0],
                self.part_advectedFields[0].gpu_data[0])
            self.init_transpose_xy(
                self.advectedFields[0].gpu_data[1],
                self.part_advectedFields[0].gpu_data[1])

    @debug
    def _apply_in_dir_vector_3d(self, simulation, dtCoeff, split_id, old_dir):
        """
        Apply operator along specified splitting direction. For 3D vector
        advection.
        @param t : Current time
        @param dt : Time step
        @param d : Splitting direction
        @param split_id : Splitting step id
        """
        # Check new data on host
        for df in self.input:
            if df.contains_data and not df.data_on_device:
                df.toDevice()
        self.cl_env.queue.finish()
        # Particle init
        if old_dir == self.dir:
            self.init_copy(self.advectedFields[0].gpu_data[0],
                           self.part_advectedFields[0].gpu_data[0])
            self.init_copy(self.advectedFields[0].gpu_data[1],
                           self.part_advectedFields[0].gpu_data[1])
            self.init_copy(self.advectedFields[0].gpu_data[2],
                           self.part_advectedFields[0].gpu_data[2])
        else:
            if min(old_dir, self.dir) == 0:
                self.init_transpose_xy(
                    self.advectedFields[0].gpu_data[0],
                    self.part_advectedFields[0].gpu_data[0])
                self.init_transpose_xy(
                    self.advectedFields[0].gpu_data[1],
                    self.part_advectedFields[0].gpu_data[1])
                self.init_transpose_xy(
                    self.advectedFields[0].gpu_data[2],
                    self.part_advectedFields[0].gpu_data[2])
            else:
                self.init_transpose_xz(
                    self.advectedFields[0].gpu_data[0],
                    self.part_advectedFields[0].gpu_data[0])
                self.init_transpose_xz(
                    self.advectedFields[0].gpu_data[1],
                    self.part_advectedFields[0].gpu_data[1])
                self.init_transpose_xz(
                    self.advectedFields[0].gpu_data[2],
                    self.part_advectedFields[0].gpu_data[2])

    @debug
    def finalize(self):
        """
        Free OpenCL memory.
        """
        for gpudf in self.variables:
            gpudf.release()
        if self.init_copy.f_timer is not None:
            for f_timer in self.init_copy.f_timer:
                self.timer.addFunctionTimer(f_timer)
            for f_timer in self.init_transpose_xy.f_timer:
                self.timer.addFunctionTimer(f_timer)
            if len(self.resolution) == 3:
                for f_timer in self.init_transpose_xz.f_timer:
                    self.timer.addFunctionTimer(f_timer)
        ParticleAdvection.finalize(self)

    def __str__(self):
        s = "GPUParticleAdvection (DiscreteOperator). "
        s += GPUParticleAdvection.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : GPUParticleAdvection"
    print GPUParticleAdvection.__doc__
