"""
@file gpu_particle_advection_1k.py

Discrete advection representation.
"""
from parmepy.constants import debug
from parmepy.gpu.gpu_particle_advection import GPUParticleAdvection
from parmepy.fields.continuous import Field
from parmepy.gpu.gpu_discrete import GPUVectorField
from parmepy.gpu.gpu_discrete import GPUScalarField
from parmepy.gpu import PARMES_REAL_GPU
from parmepy.gpu.gpu_kernel import KernelLauncher


class GPUParticleAdvection1k(GPUParticleAdvection):
    """
    Particle advection operator representation on GPU with a single kernel
    for computing advection and remeshing.
    """

    @debug
    def __init__(self, velocity, advectedFields, d,
                 part_position=None, part_advectedFields=None,
                 platform_id=0, device_id=0,
                 device_type='gpu',
                 method='', src=None, precision=PARMES_REAL_GPU):
        """
        Create a Advection operator.
        Work on a given scalar at a given velocity to produce scalar
        distribution at new positions.

        @param velocity : Velocity field
        @param advectedFields : Advected field (only single field is supported)
        @param d : Direction to advect
        @param part_position : DiscreteField to use for particle position.
        @param part_advectedFields : DiscreteField to use for particle
        advected fields.
        @param platform_id : OpenCL platform id (default = 0).
        @param device_id : OpenCL device id (default = 0).
        @param device_type : OpenCL device type (default = 'gpu').
        @param method : the method to use. {'m4prime', 'm6prime', 'm8prime'}
        @param src : User OpenCL sources.
        @param precision : Floating point precision required (default = Single
        Precision)
        """
        GPUParticleAdvection.__init__(self, velocity, advectedFields, d,
                                      part_position, part_advectedFields,
                                      platform_id, device_id,
                                      device_type,
                                      method, src,
                                      precision)
        self.name += " 1k"

    @debug
    def setUp(self):
        GPUParticleAdvection.setUp(self)

    def _buffer_allocations(self):
        """
        Allocate OpenCL buffers for velocity and advected field.
        It uses a kind of double buffering for advected field, one buffer for
        grid values and the other for particles.
        """
        ## Velocity.
        GPUVectorField.fromVectorField(self.cl_env.queue, self.velocity,
                                       self.gpu_precision)
        ## Transported scalar.
        if self.advectedFields[0].isVector:
            GPUVectorField.fromVectorField(self.cl_env.queue,
                                           self.advectedFields[0],
                                           self.gpu_precision,
                                           layout=False)
        else:
            GPUScalarField.fromScalarField(self.cl_env.queue,
                                           self.advectedFields[0],
                                           self.gpu_precision)
        ## Result scalar
        if self.part_advectedFields is None:
            self.part_advectedFields = [
                Field(self.advectedFields[0].topology.domain,
                      name="Particle_AdvectedFields",
                      isVector=self.advectedFields[0].isVector
                      ).discretize(self.advectedFields[0].topology)]
        if self.part_advectedFields[0].isVector:
            GPUVectorField.fromVectorField(
                self.cl_env.queue, self.part_advectedFields[0],
                self.gpu_precision, layout=False)
        else:
            GPUScalarField.fromScalarField(
                self.cl_env.queue, self.part_advectedFields[0],
                self.gpu_precision)

        self.variables = [self.advectedFields[0], self.velocity,
                          self.part_advectedFields[0]]
        total_mem_used = 0
        for gpuv in self.variables:
            total_mem_used += gpuv.mem_size
        print ""
        print "Total Device Global Memory used : ", total_mem_used,
        print "Bytes (", total_mem_used / (1024 ** 2), "MB)",
        print "({0:.3f} %)".format(
            100 * total_mem_used / (self.cl_env.device.global_mem_size * 1.))

    def _collect_kernels_cl_src(self):
        """
        Compile OpenCL sources for advection and remeshing kernel.

        Advection and remeshing kernel configuration:\n
        Source files:
          - kernels/advection_and_remeshing.cl
          - common.cl
          - advection/builtin.cl
        <table border>
        <tr>
           <td><b>Dimension</b></td>
           <td><b>Precision</b></td>
           <td><b>kernel sources</b></td>
           <td><b>Vector length</b></td>
           <td><b>Use Banck conflict free index</b></td>
        </tr>
        <tr>
           <td rowspan="2">3</td>
           <td>Single</td>
           <td rowspan="2">remeshing/weights_builtin.cl,
           remeshing/private.cl</td>
           <td>4</td>
           <td>No</td>
        </tr>
        <tr>
           <td>Double</td>
           <td>4</td>
           <td>Yes</td>
        </tr>
        <tr>
           <td rowspan="2">2</td>
           <td>Single</td>
           <td rowspan="2">remeshing/weights_noVec_builtin.cl,
           remeshing/basic.cl</td>
           <td>8</td>
           <td>Yes</td>
        </tr>
        <tr>
           <td>Double</td>
           <td>4</td>
           <td>Yes</td>
        </tr>
        </table>
        """
        build_options = self.build_options
        file_list = ["common.cl"]
        with_noBC = True
        vec = 4
        if len(self.resolution) == 3:
            file_list.append("remeshing/weights_builtin.cl")
            if self.advectedFields[0].isVector:
                file_list.append("remeshing/private_vector_3d.cl")
            else:
                file_list.append("remeshing/private.cl")
            if self.gpu_precision is PARMES_REAL_GPU:
                with_noBC = False
        else:
            #file_list.append("remeshing/weights_noVec_builtin.cl")
            file_list.append("remeshing/weights_noVec.cl")
            if self.advectedFields[0].isVector:
                file_list.append("remeshing/basic_vector_2d.cl")
            else:
                file_list.append("remeshing/basic.cl")
            if self.gpu_precision is PARMES_REAL_GPU:
                vec = 8
        file_list.append("advection/builtin.cl")
        if self.advectedFields[0].isVector:
            if len(self.resolution) == 3:
                file_list.append(
                    "kernels/advection_and_remeshing_vector_3d.cl")
            else:
                file_list.append(
                    "kernels/advection_and_remeshing_vector_2d.cl")
        else:
            file_list.append("kernels/advection_and_remeshing.cl")
        build_options += " -D FORMULA=" + self.method.upper()
        if with_noBC:
            build_options += " -D WITH_NOBC=1"
        WINb, gwi, lwi = self.cl_env.get_WorkItems(self.resolution, vec)
        build_options_d = " -D WI_NB=" + str(WINb)
        build_options_d += " -D PART_NB_PER_WI="
        build_options_d += str(self.resolution[0] / WINb)
        build_options_d += self.cl_env.default_sizes_constants[self.dir]
        ## Build code
        prg = self.cl_env.build_src(
            file_list, build_options + build_options_d, vec)
        self.num_advec_and_remesh = KernelLauncher(
            prg.advection_and_remeshing, self.cl_env.queue, gwi, lwi)

    @debug
    def _apply_in_dir(self, simulation, dtCoeff, split_id, old_dir):
        """
        Apply advection operator.

        @param t : current time.
        @param dt : time step.
        @param d : Direction of splitting.

        Advection algorithm:
        @li 1. Particle initialization : \n
                 - by copy scalar from grid to particles if previous splitting
                 direction equals current splitting direction.\n
                 - by transposition of scalar from grid to particle.
        @li 2. Particle advection :\n
                 - compute particle position in splitting direction as a
                 scalar. Performs a RK2 resolution of dx_p/dt = a_p.
        @li 3. Profile timings of OpenCL kernels.
        """
        GPUParticleAdvection._apply_in_dir(self, simulation, dtCoeff,
                                           split_id, old_dir)
        dt = simulation.timeStep * dtCoeff
        # Advection
        self.cl_env.queue.finish()
        self.num_advec_and_remesh(self.velocity.gpu_data[self.dir],
                                  self.part_advectedFields[0].gpu_data,
                                  self.advectedFields[0].gpu_data,
                                  self.gpu_precision(dt),
                                  self.coord_min[self.dir],
                                  self.mesh_size[self.dir])
        for df in self.output:
            df.data_on_device = True
            df.contains_data = False

    @debug
    def _apply_in_dir_vector_2d(self, simulation, dtCoeff, split_id, old_dir):
        """
        Apply advection operator for 2D vector advection.
        @param t : current time.
        @param dt : time step.
        @param d : Direction of splitting.
        @see _apply_in_dir
        """
        GPUParticleAdvection._apply_in_dir_vector_2d(self, simulation, dtCoeff,
                                                     split_id, old_dir)
        dt = simulation.timeStep * dtCoeff
        # Advection
        self.cl_env.queue.finish()
        self.num_advec_and_remesh(self.velocity.gpu_data[self.dir],
                                  self.part_advectedFields[0].gpu_data[0],
                                  self.part_advectedFields[0].gpu_data[1],
                                  self.advectedFields[0].gpu_data[0],
                                  self.advectedFields[0].gpu_data[1],
                                  self.gpu_precision(dt),
                                  self.coord_min[self.dir],
                                  self.mesh_size[self.dir])
        for df in self.output:
            df.data_on_device = True
            df.contains_data = False

    @debug
    def _apply_in_dir_vector_3d(self, simulation, dtCoeff, split_id, old_dir):
        """
        Apply advection operator for 3D vector advection.
        @param t : current time.
        @param dt : time step.
        @param d : Direction of splitting.
        @see _apply_in_dir
        """
        GPUParticleAdvection._apply_in_dir_vector_3d(self, simulation, dtCoeff,
                                                     split_id, old_dir)
        dt = simulation.timeStep * dtCoeff
        # Advection
        self.cl_env.queue.finish()
        self.num_advec_and_remesh(self.velocity.gpu_data[self.dir],
                                  self.part_advectedFields[0].gpu_data[0],
                                  self.part_advectedFields[0].gpu_data[1],
                                  self.part_advectedFields[0].gpu_data[2],
                                  self.advectedFields[0].gpu_data[0],
                                  self.advectedFields[0].gpu_data[1],
                                  self.advectedFields[0].gpu_data[2],
                                  self.gpu_precision(dt),
                                  self.coord_min[self.dir],
                                  self.mesh_size[self.dir])
        for df in self.output:
            df.data_on_device = True
            df.contains_data = False

    def finalize(self):
        if self.num_advec_and_remesh.f_timer is not None:
            for f_timer in self.num_advec_and_remesh.f_timer:
                self.timer.addFunctionTimer(f_timer)
        GPUParticleAdvection.finalize(self)

    def __str__(self):
        s = "GPUParticleAdvection1k (GPUParticleAdvection). "
        s += GPUParticleAdvection.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : GPUParticleAdvection1k"
    print GPUParticleAdvection1k.__doc__
