"""
@file parmepy.gpu.tests.test_transposition
Testing copy kernels.
"""
from parmepy.domain.box import Box
from parmepy.fields.continuous import Field
from parmepy.operator.advection import Advection
from parmepy.gpu import PARMES_REAL_GPU, cl
from parmepy.constants import ORDER, np, S_DIR
from parmepy.gpu.tools import get_opencl_environment
from parmepy.gpu.gpu_kernel import KernelListLauncher, KernelLauncher
from parmepy.gpu.gpu_discrete import GPUScalarField


def test_transposition_xy2D():
    resolution = (256, 256)
    cl_env = get_opencl_environment(0, 0, 'gpu', PARMES_REAL_GPU, resolution)
    vec = 4
    src_transpose_xy = 'kernels/transpose_xy.cl'
    build_options = ""
    build_options += cl_env.default_sizes_constants[0]
    build_options += " -D PADDING_XY=1"
    build_options += " -D TILE_DIM_XY=32 -D BLOCK_ROWS_XY=8"
    gwi = (int(resolution[0] / 4), int(resolution[1]) / 4)
    lwi = (8, 8)

    # Build code
    prg = cl_env.build_src(src_transpose_xy, build_options, vec)
    init_transpose_xy = KernelLauncher(
        prg.transpose_xy, cl_env.queue, gwi, lwi)

    data_in = np.asarray(np.random.random(resolution),
                         dtype=PARMES_REAL_GPU, order=ORDER)
    data_out = np.empty_like(data_in)
    data_out2 = np.empty_like(data_in)
    data_gpu_in = cl.Buffer(cl_env.ctx,
                            cl.mem_flags.READ_WRITE,
                            size=data_in.nbytes)
    data_gpu_out = cl.Buffer(cl_env.ctx,
                             cl.mem_flags.READ_WRITE,
                             size=data_out.nbytes)
    data_gpu_out2 = cl.Buffer(cl_env.ctx,
                              cl.mem_flags.READ_WRITE,
                              size=data_out.nbytes)
    cl.enqueue_copy(cl_env.queue, data_gpu_in, data_in)
    cl.enqueue_copy(cl_env.queue, data_gpu_out, data_out)
    cl.enqueue_copy(cl_env.queue, data_gpu_out2, data_out2)
    cl_env.queue.finish()

    init_transpose_xy(data_gpu_in, data_gpu_out)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out, data_gpu_out)
    cl_env.queue.finish()
    assert np.allclose(data_out, data_in.T)

    init_transpose_xy(data_gpu_out, data_gpu_out2)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out2, data_gpu_out2)
    cl_env.queue.finish()
    assert np.allclose(data_out2, data_in)

    data_gpu_in.release()
    data_gpu_out.release()
    data_gpu_out2.release()


def test_transposition_xy2D_rect():
    resolution = (512, 256)
    resolutionT = (256, 512)
    cl_env = get_opencl_environment(0, 0, 'gpu', PARMES_REAL_GPU, resolution)
    vec = 4
    src_transpose_xy = 'kernels/transpose_xy.cl'
    build_options = ""
    # Settings are taken from destination layout as current layout.
    # gwi is computed form input layout (appears as transposed layout)
    build_options += cl_env.default_sizes_constants[1]
    build_options += " -D TILE_DIM_XY=32 -D BLOCK_ROWS_XY=8 -D PADDING_XY=1"
    gwi = (int(resolution[0] / 4),
           int(resolution[1]) / 4)
    lwi = (8, 8)
    prg = cl_env.build_src(src_transpose_xy, build_options, vec)
    init_transpose_xy_x = KernelLauncher(prg.transpose_xy,
                                         cl_env.queue,
                                         gwi, lwi)

    build_options = ""
    build_options += cl_env.default_sizes_constants[0]
    build_options += " -D TILE_DIM_XY=32 -D BLOCK_ROWS_XY=8 -D PADDING_XY=1"
    gwi = (int(resolution[1] / 4),
           int(resolution[0]) / 4)
    lwi = (8, 8)
    prg = cl_env.build_src(src_transpose_xy, build_options, vec)
    init_transpose_xy_y = KernelLauncher(prg.transpose_xy,
                                         cl_env.queue,
                                         gwi, lwi)

    data_in = np.asarray(np.random.random(resolution),
                         dtype=PARMES_REAL_GPU, order=ORDER)
    data_out = np.empty(resolutionT,
                        dtype=PARMES_REAL_GPU, order=ORDER)
    data_out2 = np.empty(resolution,
                         dtype=PARMES_REAL_GPU, order=ORDER)
    data_gpu_in = cl.Buffer(cl_env.ctx,
                            cl.mem_flags.READ_WRITE,
                            size=data_in.nbytes)
    data_gpu_out = cl.Buffer(cl_env.ctx,
                             cl.mem_flags.READ_WRITE,
                             size=data_out.nbytes)
    data_gpu_out2 = cl.Buffer(cl_env.ctx,
                              cl.mem_flags.READ_WRITE,
                              size=data_out2.nbytes)
    cl.enqueue_copy(cl_env.queue, data_gpu_in, data_in)
    cl.enqueue_copy(cl_env.queue, data_gpu_out, data_out)
    cl.enqueue_copy(cl_env.queue, data_gpu_out2, data_out2)
    cl_env.queue.finish()

    # gpu_out <- gpu_in.T
    init_transpose_xy_x(data_gpu_in, data_gpu_out)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out, data_gpu_out)
    cl_env.queue.finish()
   # data_in_t = data_in.swapaxes(0, 1)
    assert np.allclose(data_out, data_in.T)

    # gpu_in <- gpu_out.T ( = gpu_in.T.T = gpu_in)
    init_transpose_xy_y(data_gpu_out, data_gpu_out2)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out2, data_gpu_out2)
    cl_env.queue.finish()
    assert np.allclose(data_out2, data_in)

    data_gpu_in.release()
    data_gpu_out.release()
    data_gpu_out2.release()


def test_transposition_xy3D():
    resolution = (32, 32, 32)
    cl_env = get_opencl_environment(0, 0, 'gpu', PARMES_REAL_GPU, resolution)
    vec = 2
    src_transpose_xy = 'kernels/transpose_xy.cl'
    build_options = ""
    build_options += cl_env.default_sizes_constants[0]
    build_options += " -D TILE_DIM_XY=16 -D BLOCK_ROWS_XY=8 -D PADDING_XY=1"
    gwi = (int(resolution[0] / 2),
           int(resolution[1] / 2),
           int(resolution[2]))
    lwi = (8, 8, 1)
    prg = cl_env.build_src(src_transpose_xy, build_options, vec)
    init_transpose_xy = KernelLauncher(
        prg.transpose_xy, cl_env.queue, gwi, lwi)

    data_in = np.asarray(np.random.random(resolution),
                         dtype=PARMES_REAL_GPU, order=ORDER)
    data_out = np.empty_like(data_in)
    data_out2 = np.empty_like(data_in)
    data_gpu_in = cl.Buffer(cl_env.ctx,
                            cl.mem_flags.READ_WRITE,
                            size=data_in.nbytes)
    data_gpu_out = cl.Buffer(cl_env.ctx,
                             cl.mem_flags.READ_WRITE,
                             size=data_out.nbytes)
    data_gpu_out2 = cl.Buffer(cl_env.ctx,
                              cl.mem_flags.READ_WRITE,
                              size=data_out2.nbytes)
    cl.enqueue_copy(cl_env.queue, data_gpu_in, data_in)
    cl.enqueue_copy(cl_env.queue, data_gpu_out, data_out)
    cl.enqueue_copy(cl_env.queue, data_gpu_out2, data_out2)
    cl_env.queue.finish()

    init_transpose_xy(data_gpu_in, data_gpu_out)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out, data_gpu_out)
    cl_env.queue.finish()
    assert np.allclose(data_out, data_in.swapaxes(0, 1))

    init_transpose_xy(data_gpu_out, data_gpu_out2)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out2, data_gpu_out2)
    cl_env.queue.finish()
    assert np.allclose(data_out2, data_in)

    data_gpu_in.release()
    data_gpu_out.release()
    data_gpu_out2.release()


def test_transposition_xy3D_rect():
    resolution = (32, 64, 32)
    resolutionT = (64, 32, 32)
    cl_env = get_opencl_environment(0, 0, 'gpu', PARMES_REAL_GPU, resolution)
    vec = 2
    src_transpose_xy = 'kernels/transpose_xy.cl'
    build_options = ""
    # Settings are taken from destination layout as current layout.
    # gwi is computed form input layout (appears as transposed layout)
    build_options += cl_env.default_sizes_constants[1]
    build_options += " -D TILE_DIM_XY=16 -D BLOCK_ROWS_XY=8 -D PADDING_XY=1"
    gwi = (int(resolution[0] / 2),
           int(resolution[1] / 2),
           int(resolution[2]))
    lwi = (8, 8, 1)
    prg = cl_env.build_src(src_transpose_xy, build_options, vec)
    init_transpose_xy_x = KernelLauncher(
        prg.transpose_xy, cl_env.queue, gwi, lwi)

    build_options = ""
    build_options += cl_env.default_sizes_constants[0]
    build_options += " -D TILE_DIM_XY=16 -D BLOCK_ROWS_XY=8 -D PADDING_XY=1"
    gwi = (int(resolution[1] / 2),
           int(resolution[0] / 2),
           int(resolution[2]))
    lwi = (8, 8, 1)
    prg = cl_env.build_src(src_transpose_xy, build_options, vec)
    init_transpose_xy_y = KernelLauncher(
        prg.transpose_xy, cl_env.queue, gwi, lwi)

    data_in = np.asarray(np.random.random(resolution),
                         dtype=PARMES_REAL_GPU, order=ORDER)
    data_out = np.empty(resolutionT,
                        dtype=PARMES_REAL_GPU, order=ORDER)
    data_out2 = np.empty(resolution,
                         dtype=PARMES_REAL_GPU, order=ORDER)
    data_gpu_in = cl.Buffer(cl_env.ctx,
                            cl.mem_flags.READ_WRITE,
                            size=data_in.nbytes)
    data_gpu_out = cl.Buffer(cl_env.ctx,
                             cl.mem_flags.READ_WRITE,
                             size=data_out.nbytes)
    data_gpu_out2 = cl.Buffer(cl_env.ctx,
                              cl.mem_flags.READ_WRITE,
                              size=data_out2.nbytes)
    cl.enqueue_copy(cl_env.queue, data_gpu_in, data_in)
    cl.enqueue_copy(cl_env.queue, data_gpu_out, data_out)
    cl.enqueue_copy(cl_env.queue, data_gpu_out2, data_out2)
    cl_env.queue.finish()

    init_transpose_xy_x(data_gpu_in, data_gpu_out)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out, data_gpu_out)
    cl_env.queue.finish()
    assert np.allclose(data_out, data_in.swapaxes(0, 1))

    data_out = np.empty_like(data_in)
    init_transpose_xy_y(data_gpu_out, data_gpu_out2)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out2, data_gpu_out2)
    cl_env.queue.finish()
    assert np.allclose(data_out2, data_in)

    data_gpu_in.release()
    data_gpu_out.release()
    data_gpu_out2.release()


def test_transposition_xz3D():
    resolution = (32, 32, 32)
    cl_env = get_opencl_environment(0, 0, 'gpu', PARMES_REAL_GPU, resolution)
    vec = 1
    src_transpose_xz = 'kernels/transpose_xz.cl'
    build_options = ""
    build_options += cl_env.default_sizes_constants[0]
    build_options += " -D TILE_DIM_XZ=16 -D BLOCK_ROWS_XZ=4 -D PADDING_XZ=1"
    gwi = (int(resolution[0]),
           int(resolution[1] / 4),
           int(resolution[2] / 4))
    lwi = (16, 4, 4)
    prg = cl_env.build_src(src_transpose_xz, build_options, vec)
    init_transpose_xz = KernelLauncher(
        prg.transpose_xz, cl_env.queue, gwi, lwi)

    data_in = np.asarray(np.random.random(resolution),
                         dtype=PARMES_REAL_GPU, order=ORDER)
    data_out = np.empty_like(data_in)
    data_out2 = np.empty_like(data_in)
    data_gpu_in = cl.Buffer(cl_env.ctx,
                            cl.mem_flags.READ_WRITE,
                            size=data_in.nbytes)
    data_gpu_out = cl.Buffer(cl_env.ctx,
                             cl.mem_flags.READ_WRITE,
                             size=data_out.nbytes)
    data_gpu_out2 = cl.Buffer(cl_env.ctx,
                              cl.mem_flags.READ_WRITE,
                              size=data_out2.nbytes)
    cl.enqueue_copy(cl_env.queue, data_gpu_in, data_in)
    cl.enqueue_copy(cl_env.queue, data_gpu_out, data_out)
    cl.enqueue_copy(cl_env.queue, data_gpu_out2, data_out2)
    cl_env.queue.finish()

    init_transpose_xz(data_gpu_in, data_gpu_out)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out, data_gpu_out)
    cl_env.queue.finish()
    assert np.allclose(data_out, data_in.swapaxes(0, 2))

    init_transpose_xz(data_gpu_out, data_gpu_out2)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out2, data_gpu_out2)
    cl_env.queue.finish()
    assert np.allclose(data_out2, data_in)

    data_gpu_in.release()
    data_gpu_out.release()
    data_gpu_out2.release()


def test_transposition_xz3D_rect():
    resolution = (32, 32, 64)
    resolutionT = (64, 32, 32)
    cl_env = get_opencl_environment(0, 0, 'gpu', PARMES_REAL_GPU, resolution)
    vec = 1
    src_transpose_xz = 'kernels/transpose_xz.cl'
    build_options = ""
    # Settings are taken from destination layout as current layout.
    # gwi is computed form input layout (appears as transposed layout)
    build_options += cl_env.default_sizes_constants[2]
    build_options += " -D TILE_DIM_XZ=16 -D BLOCK_ROWS_XZ=4 -D PADDING_XZ=1"
    gwi = (int(resolution[0]),
           int(resolution[1] / 4),
           int(resolution[2] / 4))
    lwi = (16, 4, 4)
    prg = cl_env.build_src(src_transpose_xz, build_options, vec)
    init_transpose_xz_x = KernelLauncher(
        prg.transpose_xz, cl_env.queue, gwi, lwi)

    build_options = ""
    build_options += cl_env.default_sizes_constants[0]
    build_options += " -D TILE_DIM_XZ=16 -D BLOCK_ROWS_XZ=4 -D PADDING_XZ=1"
    gwi = (int(resolution[2]),
           int(resolution[1] / 4),
           int(resolution[0] / 4))
    lwi = (16, 4, 4)
    prg = cl_env.build_src(src_transpose_xz, build_options, vec)
    init_transpose_xz_z = KernelLauncher(
        prg.transpose_xz, cl_env.queue, gwi, lwi)

    data_in = np.asarray(np.random.random(resolution),
                         dtype=PARMES_REAL_GPU, order=ORDER)
    data_res = data_in.copy().swapaxes(0, 2)
    data_out = np.empty_like(data_res)
    data_out2 = np.empty_like(data_in)
    data_gpu_in = cl.Buffer(cl_env.ctx,
                            cl.mem_flags.READ_WRITE,
                            size=data_in.nbytes)
    data_gpu_out = cl.Buffer(cl_env.ctx,
                             cl.mem_flags.READ_WRITE,
                             size=data_out.nbytes)
    data_gpu_out2 = cl.Buffer(cl_env.ctx,
                              cl.mem_flags.READ_WRITE,
                              size=data_out2.nbytes)
    cl.enqueue_copy(cl_env.queue, data_gpu_in, data_in)
    cl.enqueue_copy(cl_env.queue, data_gpu_out, data_out)
    cl.enqueue_copy(cl_env.queue, data_gpu_out2, data_out2)
    cl_env.queue.finish()

    init_transpose_xz_x(data_gpu_in, data_gpu_out)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out, data_gpu_out)
    cl_env.queue.finish()
    assert np.allclose(data_out, data_res)

    init_transpose_xz_z(data_gpu_out, data_gpu_out2)
    cl_env.queue.finish()
    cl.enqueue_copy(cl_env.queue, data_out2, data_gpu_out2)
    cl_env.queue.finish()
    assert np.allclose(data_out2, data_in)

    data_gpu_in.release()
    data_gpu_out.release()
    data_gpu_out2.release()



if __name__ == '__main__':
    test_transposition_xz3D_rect()
