# -*- codingind utf-8 -*-
"""
@file differential_operations.py
Library of functions used to perform classical vector calculus
(diff operations like grad, curl ...)
"""
from parmepy.constants import debug, XDIR, YDIR, ZDIR
from abc import ABCMeta, abstractmethod
from finite_differences import FD_C_4


class DifferentialOperation(object):
    __metaclass__ = ABCMeta

    ## @debug
    ## def __new__(cls, *args, **kw):
    ##     return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self):
        pass


class Curl(DifferentialOperation):
    """
    Computes \f$ \nabla \ times (f) \f$, f being a vector field.
    """
    def __init__(self, topo, method=None):
        if method is not None and method.find('FD_order2') >= 0:
                raise ValueError("2nd order scheme Not yet implemented")
        else:
            assert (topo.ghosts >= 2).all(),\
                'you need a ghost layer for FD4 scheme.'
            self.fcall = self.FDCentral4
            self.fd_scheme = FD_C_4(topo.mesh.space_step)
            self.indices = topo.mesh.iCompute
            self.fd_scheme.computeIndices(self.indices)

    def __call__(self, variable, data=None):
        return self.fcall(variable, data)

    def FDCentral4(self, variable, data=None):
        """
        """
        # Get slices for subarrays ...
       # init (if required) return data
        if data is None:
            data = []
            for i in xrange(variable.dimension):
                data.append(variable.data[i].copy())
                #data[i][...] = 0.0

        #data[0][...] = 0.0
        #data[1][...] = 0.0
        #data[2][...] = 0.0

        #--  d/dy vz -- in data[XDIR]
        self.fd_scheme.inplace(variable[ZDIR], YDIR, data[XDIR])
        # -- -d/dz vy -- in data[YDIR]
        work = data[YDIR]
        self.fd_scheme.inplace(variable[YDIR], ZDIR, work)
        # data_x = d/dy vz - d/dz vy
        data[XDIR][self.indices] -= work[self.indices]

        #--  d/dz vx -- in data[YDIR]
        self.fd_scheme.inplace(variable[XDIR], ZDIR, data[YDIR])
        # -- -d/dx vz -- in data[ZDIR]
        work = data[ZDIR]
        self.fd_scheme.inplace(variable[ZDIR], XDIR, work)

        # data_y = d/dz vx - d/dx vz
        data[YDIR][self.indices] -= work[self.indices]

        # - d/dy vx in data[ZDIR]
        self.fd_scheme.inplace(variable[XDIR], YDIR, data[ZDIR])
        data[ZDIR][self.indices] *= -1.
        # add d/dx vy, add to data[ZDIR]
        self.fd_scheme.accumulate(variable[YDIR], XDIR, data[ZDIR])

        return data


class DivV(DifferentialOperation):
    """
    Computes \f$ \nabla.(\rho V) \f$, \f$ \rho\f$ a scalar field and
    V a vector field.
    """
    def __init__(self, topo, method=None):
        if method is not None and method.find('FD_order2') >= 0:
                raise ValueError("2nd order scheme Not yet implemented")
        elif method.find('FD_C4_nowork') >= 0:
            self.fcall = self.FDCentral4
            self.fd_scheme = FD_C_4((topo.mesh.space_step))
        else:
            self.fcall = self.FDCentral4Work
            self.fd_scheme = FD_C_4((topo.mesh.space_step))

        self.indices = topo.mesh.iCompute
        self.fd_scheme.computeIndices(self.indices)
        assert (topo.ghosts >= 2).all(),\
            'you need a ghost layer for FD4 scheme.'

    def __call__(self, var1, scal, data=None, work=None):
        return self.fcall(var1, scal, data)

    def FDCentral4Work(self, var1, scal, data=None, work=None):
        """
        Compute central finite difference scheme, order 4.
        """
        if data is None:
            data = var1.data[0].copy()
        if work is None:
            work = var1.data[0].copy()

        #data[...] = 0.0

        #  d/dx (scal * v1x), saved in data
        self.fd_scheme.inplace(scal * var1[XDIR], XDIR, data)
        #  d/dy (scal * v1y), saved in work
        self.fd_scheme.inplace(scal * var1[YDIR], YDIR, work)
        data[self.indices] += work[self.indices]
        #  d/dz(scal * v1z), saved in work
        self.fd_scheme.inplace(scal * var1[ZDIR], ZDIR, work)
        data[self.indices] += work[self.indices]
        return data

    def FDCentral4(self, var1, scal, data=None):
        """
        """
        if data is None:
            data = var1.data[0].copy()

        data[...] = 0.0

        #  d/dx (scal * v1x)
        self.fd_scheme.accumulate(var1[XDIR] * scal, XDIR, data)
        #  d/dy (scal * v1y)
        self.fd_scheme.accumulate(var1[YDIR] * scal, YDIR, data)
        #  d/dz(scal * v1z)
        self.fd_scheme.accumulate(var1[ZDIR] * scal, ZDIR, data)

        return data


class DivT(DifferentialOperation):
    """
    Computes \f$ \nabla.(wx.V,wy.V,wz.V) \f$, \f$w\f$ and V some vector fields.
    """
    def __init__(self, topo, method=None):
        if method is not None and method.find('FD_order2') >= 0:
                raise ValueError("2nd order scheme Not yet implemented")
        else:
            self.fcall = self.FDCentral4
            self.div = DivV(topo, method='FD_C4')
            self.divZ = DivV(topo, method='FD_C4_nowork')

    def __call__(self, var1, var2, data=None):
        return self.fcall(var1, var2, data)

    def FDCentral4(self, var1, var2, data=None):
        """
        """
        # init (if required) return data
        if data is None:
            data = []
            for i in xrange(var1.dimension):
                data.append(var1.data[i].copy())
        # we use data[Y] and data[Z] as work ...
        data[XDIR] = self.div(var1, var2[XDIR],
                              data=data[XDIR], work=data[YDIR])
        data[YDIR] = self.div(var1, var2[YDIR],
                              data=data[YDIR], work=data[ZDIR])
        # No work vector for zdir
        data[ZDIR] = self.divZ(var1, var2[ZDIR], data=data[ZDIR])
        return data
