"""
@file finite_differences.py
Finite difference schemes description
"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug, PARMES_REAL
import numpy as np


class FiniteDifference(object):
    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @abstractmethod
    def __init__(self, *args):
        """
        Build a finite difference scheme
        """


class FD_C_4(FiniteDifference):
    """
    Centered scheme, 4th order.
    """

    def __init__(self, step):
        """
        Constructor
        @param step : grid step size
        """
        #  dim of the field on which scheme will be applied
        # (i.e dim of the domain)
        self._dim = step.size
        # list of indices for index - 1
        self._m1 = []
        # for index - 2
        self._m2 = []
        # index + 1
        self._a1 = []
        # index + 2
        self._a2 = []
        # FD scheme coefficients
        self._coeff = np.asarray(1. / (12. * step), dtype=PARMES_REAL)

    def computeIndices(self, indices):

        self.indices = indices
        self._m1 = []
        self._m2 = []
        self._a1 = []
        self._a2 = []
        for dim in xrange(self._dim):
            self._m1.append(list(self.indices))
            self._m1[dim][dim] = slice(self.indices[dim].start - 1,
                                       self.indices[dim].stop - 1,
                                       self.indices[dim].step)
            self._m2.append(list(self.indices))
            self._m2[dim][dim] = slice(self.indices[dim].start - 2,
                                       self.indices[dim].stop - 2,
                                       self.indices[dim].step)
            self._a1.append(list(self.indices))
            self._a1[dim][dim] = slice(self.indices[dim].start + 1,
                                       self.indices[dim].stop + 1,
                                       self.indices[dim].step)
            self._a2.append(list(self.indices))
            self._a2[dim][dim] = slice(self.indices[dim].start + 2,
                                       self.indices[dim].stop + 2,
                                       self.indices[dim].step)
        #return self._m2, self._m1, self._a1, self._a2

    def inplace(self, tab, cdir, result):
        """
        Compute \f$ \frac{\partial d tab()}{\partial cdir} \f$ using fd scheme,
        The result is saved in input/ouput parameter result (in place!)
        @param tab : a numpy array
        @param cdir : direction of differentiation
        @param result : input/output numpy array to save the resulting data.
        """
        result[self.indices] = tab[self._a1[cdir]]
        result[self.indices] -= tab[self._m1[cdir]]
        result[self.indices] *= 8
        result[self.indices] += tab[self._m2[cdir]]
        result[self.indices] -= tab[self._a2[cdir]]
        result[self.indices] *= self._coeff[cdir]

    def accumulate(self, tab, cdir, result):
        """
        Compute \f$ \frac{\partial d tab()}{\partial cdir} \f$ using fd scheme,
        The result is ADDED to input/ouput parameter result.
        @param tab : a numpy array
        @param cdir : direction of differentiation
        @param result : input/output numpy array to save the resulting data.
        """
        result[self.indices] += self._coeff[cdir] * (8 *
                                                     (tab[self._a1[cdir]] -
                                                      tab[self._m1[cdir]]) +
                                                     tab[self._m2[cdir]] -
                                                     tab[self._a2[cdir]])
