"""
@file remeshing.py
"""
from parmepy.constants import np
from parmepy.numerics.method import NumMethod


class Remeshing(NumMethod):
    """Remshing"""

    def __init__(self, dim, dx, origin, formula):
        """
        Create a remeshing numeric method based on given formula.
        @param dim : problem dimension
        @param dx : mesh space step
        @param origin : mesh lower point
        @param formula : Remeshing formula to use.
        Availables formulas :
          - M4prime
          - M6prime
          - M8prime
          - L6star
        """
        NumMethod.__init__(self)
        self.name = formula
        self.dx = dx
        self.origin = origin
        self._slice_all = [slice(None, None, None)
                           for d in xrange(dim)]
        if self.name == 'm4prime':
            self._shift = 1
            self._weights = [
                lambda y, s: s * ((y * (y * (-y + 2.) - 1.)) / 2.),
                lambda y, s: s * ((y * y * (3. * y - 5.) + 2.) / 2.),
                lambda y, s: s * ((y * (y * (-3. * y + 4.) + 1.)) / 2.),
                lambda y, s: s * ((y * y * (y - 1.)) / 2.)
                ]
        elif self.name == 'l4star':
            self._shift = 2
            self._weights = [
                lambda y, s: s * y * (y * (y * (y * (13. - 5. * y) - 9.) - 1.) + 2.) / 24.,
                lambda y, s: s * y * (y * (y * (y * (25. * y - 64.) + 39.) + 16.) - 16.) / 24.,
                lambda y, s: s * (y * y * (y * (y * (126. - 50. * y) - 70.) - 30.) / 24. + 1.),
                lambda y, s: s * y * (y * (y * (y * (50. * y - 124.) + 66.) + 16.) + 16.) / 24.,
                lambda y, s: s * y * (y * (y * (y * (61. - 25. * y) - 33.) - 1.) - 2.) / 24.,
                lambda y, s: s * y * y * y * (y * (5. * y - 12.) + 7.) / 24.
                ]
        elif self.name == 'l4star_c3':
            self._shift = 2
            self._weights = [
                lambda y, s: s * (2. + (-1. + (-2. + (-22. + (58. + (-49. + 14. * y) * y) * y) * y) * y) * y) * y / 24.,
                lambda y, s: s * (-16. + (16. + (4. + (111. + (-290. + (245. - 70. * y) * y) * y) * y) * y) * y) * y / 24.,
                lambda y, s: s * (1. + (-30. + (-224. + (580. + (-490. + 140. * y) * y) * y) * y * y) * y * y / 24.),
                lambda y, s: s * (16. + (16. + (-4. + (226. + (-580. + (490. - 140. * y) * y) * y) * y) * y) * y) * y / 24.,
                lambda y, s: s * (-2. + (-1. + (2. + (-114. + (290. + (-245. + 70. * y) * y) * y) * y) * y) * y) * y / 24.,
                lambda y, s: s * (23. + (-58. + (49. - 14. * y) * y) * y) * y * y * y * y / 24.

                ]
        elif self.name == 'l4star_c4':
            self._shift = 2
            self._weights = [
                lambda y, s: s * (2. + (-1. + (-2. + (1. + (-80. + (273. + (-354. + (207. - 46. * y) * y) * y) * y) * y) * y) * y) * y) * y / 24.,
                lambda y, s: s * (-16. + (16. + (4. + (-4. + (400. + (-1365. + (1770. + (-1035. + 230. * y) * y) * y) * y) * y) * y) * y) * y) * y / 24.,
                lambda y, s: s * (1. + (-30. + (6. + (-800. + (2730. + (-3540. + (2070. - 460. * y) * y) * y) * y) * y) * y * y) * y * y / 24.),
                lambda y, s: s * (16. + (16. + (-4. + (-4. + (800. + (-2730. + (3540. + (-2070. + 460. * y) * y) * y) * y) * y) * y) * y) * y) * y / 24.,
                lambda y, s: s * (-2. + (-1. + (2. + (1. + (-400. + (1365. + (-1770. + (1035. - 230. * y) * y) * y) * y) * y) * y) * y) * y) * y / 24.,
                lambda y, s: s * (80. + (-273. + (354. + (-207. + 46. * y) * y) * y) * y) * y * y * y * y * y / 24.
                ]
        elif self.name == 'm8prime':
            self._shift = 3
            self._weights = [
                lambda y, s: s * ((y * (y * (y * (y * (y * (y * (-10. * y + 21.) + 28.) - 105.)+ 70.) + 35.) - 56.) + 17.) / 3360.),
                lambda y, s: s * ((y * (y * (y * (y * (y * (y * (70. * y - 175.) - 140.) + 770.)- 560.) - 350.) + 504.) - 102.) / 3360.),
                lambda y, s: s * ((y * (y * (y * (y * (y * (y * (-210. * y + 609.) + 224.) - 2135.)+ 910.) + 2765.) - 2520.) + 255.) / 3360.),
                lambda y, s: s * ((y * y * (y * y * (y * y * (70. * y - 231.) + 588.) - 980.) + 604.)/ 672.),
                lambda y, s: s * ((y * (y * (y * (y * (y * (y * (-70. * y + 259.) - 84.) - 427.)- 182.) + 553.) + 504.) + 51.) / 672.),
                lambda y, s: s * ((y * (y * (y * (y * (y * (y * (210. * y - 861.) + 532.) + 770.)+ 560.) - 350.) - 504.) - 102.) / 3360.),
                lambda y, s: s * ((y * (y * (y * (y * (y * (y * (-70. * y + 315.) - 280.) - 105.)- 70.) + 35.) + 56.) + 17.) / 3360.),
                lambda y, s: s * ((y * y * y * y * y * (y * (10. * y - 49.) + 56.)) / 3360.)
                ]
        elif self.name == 'l6star':
            self._shift = 3
            self._weights = [
                lambda y, s: s * (((-12. + (4. + (15. + (140. + (-370.+ (312. - 89. * y) * y) * y) * y) * y) * y) * y) / 720.),
                lambda y, s: s * (((108. + (-54. + (-120. + (-955. +(2581. + (-2183. + 623. * y) * y) * y) * y) * y) * y ) * y) / 720.),
                lambda y, s: s * (((-180. + (180. + (65. + (950. +(-2574. + (2182. - 623. * y) * y) * y) * y) * y) * y) * y) / 240.),
                lambda y, s: s * ( 1. + ((-196. + (-959. + (2569. +(-2181. + 623. * y) * y) * y) * y * y ) * y * y) / 144.),
                lambda y, s: s * (((108. + (108. + (-39. + (976. + (-2566.+ (2180. - 623. * y) * y) * y) * y) * y) * y) * y) / 144.),
                lambda y, s: s * (((-36. + (-18. + (40. + (-995. + (2565.+ (-2179. + 623. * y) * y) * y) * y) * y) * y) * y) / 240.),
                lambda y, s: s * (((12. + (4. + (-15. + (1010. + (-2566.+ (2178. - 623. * y) * y) * y) * y) * y) * y) * y) / 720.),
                lambda y, s: s * (((-145. + (367. + (-311. + 89. * y)* y) * y) * y * y * y * y) / 720.)
                ]
        elif self.name == 'l6star_c4':
            self._shift = 3
            self._weights = [
                lambda y, s: s * (-12. + (4. + (15. + (-5. + (500. + (-1718. + (2231. + (-1305. + 290. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (108. + (-54. + (-120. + (60. + (-3509. + (12027. + (-15617. + (9135. - 2030. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (-540. + (540. + (195. + (-195. + (10548. + (-36084. + (46851. + (-27405. + 6090. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (1. + (-980. + (280. + (-17605. + (60145. + (-78085. + (45675. - 10150. * y) * y) * y) * y) * y) * y * y) * y * y / 720.),
                lambda y, s: s * (540. + (540. + (-195. + (-195. + (17620. + (-60150. + (78085. + (-45675. + 10150. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (-108. + (-54. + (120. + (60. + (-10575. + (36093. + (-46851. + (27405. - 6090. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (12. + (4. + (-15. + (-5. + (3524. + (-12032. + (15617. + (-9135. + 2030. * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (-503. + (1719. + (-2231. + (1305. - 290. * y) * y) * y) * y) * y * y * y * y * y / 720.
                ]
        elif self.name == 'l6star_c5':
            self._shift = 3
            self._weights = [
                lambda y, s: s * (-12. + (4. + (15. + (-5. + (-3. + (1803. + (-7829. + (13785. + (-12285. + (5533. - 1006. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (108. + (-54. + (-120. + (60. + (12. + (-12620. + (54803. + (-96495. + (85995. + (-38731. + 7042. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (-540. + (540. + (195. + (-195. + (-15. + (37857. + (-164409. + (289485. + (-257985. + (116193. - 21126. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (1. + (-980. + (280. + (-63090. + (274015. + (-482475. + (429975. + (-193655. + 35210. * y) * y) * y) * y) * y) * y * y) * y * y) * y * y / 720.),
                lambda y, s: s * (540. + (540. + (-195. + (-195. + (15. + (63085. + (-274015. + (482475. + (-429975. + (193655. - 35210. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (-108. + (-54. + (120. + (60. + (-12. + (-37848. + (164409. + (-289485. + (257985. + (-116193. + 21126. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (12. + (4. + (-15. + (-5. + (3. + (12615. + (-54803. + (96495. + (-85995. + (38731. - 7042. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (-1802. + (7829. + (-13785. + (12285. + (-5533. + 1006. * y) * y) * y) * y) * y) * y * y * y * y * y * y / 720.
                ]
        elif self.name == 'l6star_c6':
            self._shift = 3
            self._weights = [
                lambda y, s: s * (-12. + (4. + (15. + (-5. + (-3. + (1. + (6587. + (-34869. + (77815. + (-93577. + (63866. + (-23426. + 3604. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (108. + (-54. + (-120. + (60. + (12. + (-6. + (-46109. + (244083. + (-544705. + (655039. + (-447062. + (163982. - 25228. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (-540. + (540. + (195. + (-195. + (-15. + (15. + (138327. + (-732249. + (1634115. + (-1965117. + (1341186. + (-491946. + 75684. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (1. + (-980. + (280. + (-20. + (-230545. + (1220415. + (-2723525. + (3275195. + (-2235310. + (819910. - 126140. * y) * y) * y) * y) * y) * y) * y) * y * y) * y * y) * y * y / 720.),
                lambda y, s: s * (540. + (540. + (-195. + (-195. + (15. + (15. + (230545. + (-1220415. + (2723525. + (-3275195. + (2235310. + (-819910. + 126140. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (-108. + (-54. + (120. + (60. + (-12. + (-6. + (-138327. + (732249. + (-1634115. + (1965117. + (-1341186. + (491946. - 75684. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (12. + (4. + (-15. + (-5. + (3. + (1. + (46109. + (-244083. + (544705. + (-655039. + (447062. + (-163982. + 25228. * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y) * y / 720.,
                lambda y, s: s * (-6587. + (34869. + (-77815. + (93577. + (-63866. + (23426. - 3604. * y) * y) * y) * y) * y) * y) * y * y * y * y * y * y * y / 720.
                ]
        else:
            raise ValueError("Unknown formula : " + self.name)

    def slice_i_along_d(self, i, d):
        l = list(self._slice_all)
        l[d] = i
        return tuple(l)

    def __call__(self, ppos, origin, pscal, d, res=None,
                 i_row=None, i_col=None, i_dep=None, i_y=None, work=None):

        """
        Remesh particles at position ppos with scalar
        pscal along direction d.

        @param ppos : particle position
        @param pscal : particle scalar
        @param d : splitting direction
        @param i_row : allocated working array (rows indices array).
        @param i_col : allocated working array (columns indices array).
        @param i_dep : allocated working array (depth indices array).
        @param i_y : allocated working array (distance to left grid point).
        """
        shape = ppos.shape
        if len(shape) == 2:
            index = tuple([i_row, i_col])
            if d == 0:
                index[1][...] = np.indices((shape[1],))[0][np.newaxis, :]
            else:
                index[0][...] = np.indices((shape[0],))[0][:, np.newaxis]
        elif len(shape) == 3:
            index = tuple([i_row, i_col, i_dep])
            if d == 0:
                index[1][...] = \
                    np.indices((shape[1],))[0][np.newaxis, :, np.newaxis]
                index[2][...] = \
                    np.indices((shape[2],))[0][np.newaxis, np.newaxis, :]
            elif d == 1:
                index[0][...] = \
                    np.indices((shape[0],))[0][:, np.newaxis, np.newaxis]
                index[2][...] = \
                    np.indices((shape[2],))[0][np.newaxis, np.newaxis, :]
            else:
                index[0][...] = \
                    np.indices((shape[0],))[0][:, np.newaxis, np.newaxis]
                index[1][...] = \
                    np.indices((shape[1],))[0][np.newaxis, :, np.newaxis]
        else:
            index = tuple([i_row, ])

        floor = res
        floor[...] = (ppos - origin) / self.dx[d]
        i_y[...] = (floor)
        floor[...] = np.floor(floor)
        i_y[...] -= floor

        index[d][...] = (floor.astype(np.int32) - self._shift) % shape[d]
        res[...] = 0.

        for w_id, w in enumerate(self._weights):
            if w_id > 0:
                index[d][...] = (index[d] + 1) % shape[d]
            for i in xrange(shape[d]):
                sl = self.slice_i_along_d(i, d)
                work = w(i_y, pscal)
                index_sl = tuple([ind[sl] for ind in index])
                res[index_sl] = res[index_sl] + work[sl]
