"""
@file operator/continuous.py

Interface common to all continuous operators.
"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug
from parmepy.mpi import main_rank
import numpy as np


class Operator(object):
    """
    Abstract base class for continuous operators.

    An operator is composed of :
    - a set of continuous variables (at least one)
    - a method which defined how it would be discretized/processed
    - a discrete operator : object build using the method
    and the discretized variables.

    To each variable a 'resolution' is associated, used
    to create a topology and a discrete field.
    See details in 'real' operators (derived classes) description.
    """
    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, variables, method=None, name=None,
                 topo=None, ghosts=None):
        """
        Build the operator.
        The only required parameter is a list of variables.
        @param variables : list of fields on which this operator will apply.
        @param method : name of the method used to descretize
        the operator.
        @param name : operator id.
        @param topo : a predefined topology to discretize variables
        @param ghosts : number of points in the ghost layer
        """
        ## List of parmepy.continuous.Fields involved in the operator.
        if isinstance(variables, list):
            self.variables = variables
        else:
            self.variables = [variables]
        ## Input variables.
        self.input = []
        ## Output variables.
        self.output = []
        ## Domain of definition.
        self.domain = self.variables[0].domain
        for v in self.variables:
            assert v.domain is self.domain, 'All variables of the operator\
                must be defined on the same domain.'
        ## Object to describe the method of discretization of this operator.
        self.discreteOperator = None
        ## Optional name to identify the operator.
        self.name = name
        ## bool to check if the setup function has been called for
        ## this operator
        self._isUpToDate = False
        ## A dictionnary of discreteFields associated with this operator
        ## key = continuous variable \n
        ## Example : discrete_variable = discreteFields[velocity]
        ## returns the discrete fields that corresponds to the continuous
        ## variable velocity.
        self.discreteFields = {}
        ## The method used to discretize the operator.
        self.method = method
        ## Number of points in the ghost layer
        if ghosts is not None:
            self.ghosts = np.asarray(ghosts)
            print "ooioii", self.ghosts, self.ghosts.size, self.domain.dimension
            assert self.ghosts.size == self.domain.dimension
        else:
            self.ghosts = None

        ## A predefined topology to discretize variables.
        self._predefinedTopo = None
        if topo is not None:
            self._predefinedTopo = topo
            if self.ghosts is not None:
                assert (self.ghosts == topo.ghosts), 'topo.ghosts and\
                input for ghosts must not be different'
            else:
                self.ghosts = topo.ghosts

    @abstractmethod
    def setUp(self):
        """
        Initialize the operator : compute or get topologies,
        discretize variables, create a discrete operator.
        """

    @debug
    def finalize(self):
        """
        Memory cleaning.
        """
        self.discreteOperator.finalize()
        self.timer = self.discreteOperator.timer

    @debug
    def apply(self, simulation=None):
        """
        Apply this operator to its variables.
        @param simulation : object that describes the simulation
        parameters (time, time step, iteration number ...), see
        parmepy.problem.simulation.Simulation for details.
        """
        self.discreteOperator.apply(simulation)

    def printComputeTime(self):
        """ Time monitoring."""
        if self.discreteOperator is not None:
            self.discreteOperator.printComputeTime()
            self.time_info = self.discreteOperator.time_info
        else:
            shortName = str(self.__class__).rpartition('.')[-1][0:-2]
            s = '[' + str(main_rank) + '] ' + shortName
            s += " : operator not discretized --> no computation, time = 0."

    def isUp(self):
        """
        True if ready to be applied (--> setup function has
        been called succesfully)
        """
        return self._isUpToDate

    def __str__(self):
        """
        Common printings for operators
        """
        shortName = str(self.__class__).rpartition('.')[-1][0:-2]
        if self.discreteOperator is not None:
            s = str(self.discreteOperator)
        else:
            s = shortName + " operator. Not discretised."
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Operator"
    print Operator.__doc__
