# -*- coding: utf-8 -*-
"""
@file scales_advection.py
Discrete Advection operator based on scales library (Jean-Baptiste)

"""
from parmepy.f2py import scales2py
from discrete import DiscreteOperator
from parmepy.constants import debug, np
from parmepy.tools.timers import timed_function


class ScalesAdvection(DiscreteOperator):
    """
    Advection process, based on scales library.
    """

    @debug
    def __init__(self, velocity, advectedFields, method=''):
        """
        Constructor.
        @param velocity discrete field
        @param advectedFields : list of discrete fields to be advected
        @param method : Method used.
        """
        ## Advection velocity
        self.velocity = velocity
        variables = [self.velocity]
        if not isinstance(advectedFields, list):
            self.advectedFields = [advectedFields]
        else:
            self.advectedFields = advectedFields
        # update the list of discrete variables for this operator
        [variables.append(advecF) for advecF in self.advectedFields]

        DiscreteOperator.__init__(self, variables, method,
                                  name="SCALES Advection")

        self.input = [self.velocity]
        self.output = list(variables).remove(self.velocity)

    @debug
    @timed_function
    def apply(self, simulation=None):
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        dt = simulation.timeStep

#        for adF in self.advectedFields:
#            if isinstance(adF, VectorField):
#                for d in xrange(adF.dimension):
#                    adF[d][...] = scales2py.solve_advection(
#                        dt, self.velocity.data[0],
#                        self.velocity.data[1],
#                        self.velocity.data[2],
#                        adF[d][...])
#            else:
#                adF[...] = scales2py.solve_advection(
#                    dt, self.velocity.data[0],
#                    self.velocity.data[1],
#                    self.velocity.data[2],
#                    adF.data)


        ## Space discretization of grad(u)
        mesh_size = self.velocity.topology.mesh.space_step
        ind=np.arange(self.velocity.topology.mesh.resolution[0])

###      First order scheme
###       X components of temp and result
        temp1 = (self.velocity[0][...] -
                 self.velocity[0][np.roll(ind,+1),...]) / (mesh_size[0]) 

        temp2 = (self.velocity[0][...] -
                 self.velocity[0][:,np.roll(ind,+1),:]) / (mesh_size[1]) 

        temp3 = (self.velocity[0][...] -
                 self.velocity[0][...,np.roll(ind,+1)]) / (mesh_size[2]) 

        maxstr1= np.max(abs(temp1)+abs(temp2)+abs(temp3))
        maxadv1= np.max(abs(temp1))

#       Y components of temp and result
        temp4 = (self.velocity[1][...] -
                 self.velocity[1][np.roll(ind,+1),...]) / (mesh_size[0]) 

        temp5 = (self.velocity[1][...] -
                 self.velocity[1][:,np.roll(ind,+1),:]) / (mesh_size[1]) 

        temp6 = (self.velocity[1][...] -
                 self.velocity[1][...,np.roll(ind,+1)]) / (mesh_size[2]) 

        maxstr2= np.max(abs(temp4)+abs(temp5)+abs(temp6))
        maxadv2= np.max(abs(temp5))

#       Z components of temp and result
        temp7 = (self.velocity[2][...] -
                 self.velocity[2][np.roll(ind,+1),...]) / (mesh_size[0]) 

        temp8 = (self.velocity[2][...] -
                 self.velocity[2][:,np.roll(ind,+1),:]) / (mesh_size[1]) 

        temp9 = (self.velocity[2][...] -
                 self.velocity[2][...,np.roll(ind,+1)]) / (mesh_size[2]) 

        maxstr3= np.max(abs(temp7)+abs(temp8)+abs(temp9))
        maxadv3= np.max(abs(temp9))
        maxAdv = max(maxadv1,maxadv2,maxadv3)

        stabcoeff = 1.0 / 2.0

        print "dtAdvec", stabcoeff / maxAdv

        ## Time Subcycles
        dtadv = dt
        dtadv = min(dtadv, stabcoeff / maxAdv)
        if dt == dtadv:
            print "dtadv, ndt, stab/max ", dtadv, int(dt / dtadv),\
                  stabcoeff / maxAdv
            ndt = int(dt / dtadv)
        else:
            print "dtadv, ndt", dtadv, int(dt / dtadv) + 1, stabcoeff / maxAdv
            ndt = int(dt / dtadv) + 1
        dtadv = dt / float(ndt)

        for i in range(ndt):
            for adF in self.advectedFields:
                for d in xrange(adF.nbComponents):
                    adF[d][...] = scales2py.solve_advection(
                        dtadv, self.velocity.data[0],
                        self.velocity.data[1],
                        self.velocity.data[2],
                        adF[d][...])

    def finalize(self):
        """
        \todo check memory deallocation in scales???
        """
        DiscreteOperator.finalize(self)

if __name__ == "__main__":
    print __doc__
    print "- Provided class : ScalesAdvection"
    print ScalesAdvection.__doc__
