# -*- coding: utf-8 -*-
"""
@file operator/discrete/stretching.py

Discrete stretching representation
"""

from parmepy.constants import debug
from discrete import DiscreteOperator
from parmepy.numerics.differentialOperator import DifferentialOperator
from parmepy.numerics.fct2op import Fct2Op
from parmepy.numerics.integrators.euler import Euler
from parmepy.numerics.integrators.runge_kutta2 import RK2
from parmepy.numerics.integrators.runge_kutta3 import RK3
from parmepy.numerics.integrators.runge_kutta4 import RK4
from parmepy.tools.timers import timed_function
from parmepy.operator.discrete.synchronizeGhosts import SynchronizeGhosts_d


class Stretching_d(DiscreteOperator):
    """
    Stretching operator representation.
    """

    @debug
    def __init__(self, velocity, vorticity,
                 method='FD_order4 RK3', propertyOp='divConservation'):
        """
        Constructor.
        Create a Stretching operator on a given continuous domain.
        Work on a given velocity and vorticity to compute the stretching term.

        @param stretch : Stretching operator.
        @param method : Method to use
        """
        DiscreteOperator.__init__(self, [velocity, vorticity], method,
                                  name="Stretching")
        ## Velocity.
        self.velocity = velocity
        ## Vorticity.
        self.vorticity = vorticity
        #        ## input fields
        #        self.input = [self.velocity, self.vorticity]
        self.propertyOp = propertyOp
        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]

        self.cststretch = 0.
        if self.method.find('Euler') >= 0:
            self.num_method = Euler
            self.cststretch = 2.0
        elif self.method.find('RK2') >= 0:
            self.num_method = RK2
            self.cststretch = 2.0
        elif self.method.find('RK3') >= 0:
            self.num_method = RK3
            self.cststretch = 2.5127
        elif self.method.find('RK4') >= 0:
            self.num_method = RK4
            self.cststretch = 2.7853
        else:
            self.num_method = RK3
            self.cststretch = 2.5127

    @timed_function
    def apply(self, simulation):
        """
        Apply Stretching operator.

        Stretching algorithm:
        @li 1. discretization of the divergence operator (div wu)
        or w grad(u) : \n
        @li 2. Time integration. Performs a method choose to resolove
        the equation dw/dt = div(wu) or dw/dt = w grad(u).
         -define the function of ODEsolver as the result of the first step
         -integrate with the chosen method (RK4. RK2. euler)
        """
        if simulation is None:
            raise ValueError("Missing simulation value for computation.")

        dt = simulation.timeStep
        if self.num_method is not None:
            condA = self.propertyOp.find('divConservation') < 0
            if condA and self.propertyOp.find('massConservation') < 0:
                print 'No or wrong stretching operator property specified, \
                use default divConservation'
                self.propertyOp = 'divConservation'

            if self.propertyOp.find('massConservation') >= 0:
                ## Determination of Stretching time step (stability condition)
                self.ndt = 1
                self.dtstretch = dt

                ## fct2Op
                self.function = Fct2Op(self.vorticity,
                                       self.velocity,
                                       choice='divWU',
                                       topology=self.velocity.topology)

            else:  # self.propertyOp == 'divConservation'

                ## Space discretization of grad(u)
                self.stretchOp = DifferentialOperator(
                    self.vorticity, self.velocity,
                    method='FD_order4',
                    choice='gradV')
                #une synchro
                self.synchro = SynchronizeGhosts_d(
                    self.vorticity.data,
                    self.vorticity.topology, 'greaterSend')
                self.synchro.setUp()
                gradResult, maxArray, divergence = self.stretchOp(self.synchro)

                print "maxadv, maxgersh", maxArray[1], maxArray[0]
                print "dtAdvec", 0.5 / maxArray[1]

                ## Determination of Stretching time step (stability condition)
                self.ndt = self.stabilityTest(dt, maxArray[0],
                                              self.num_method, self.cststretch)
                self.dtstretch = dt / float(self.ndt)

                ## fct2Op
                self.function = Fct2Op(self.vorticity,
                                       gradResult,
                                       method='FD_order4',
                                       choice='gradV',
                                       topology=self.vorticity.topology)
                self.synchro.apply()

            ## Time integration
            t = simulation.time
            for i in range(self.ndt):
                methodInt = self.num_method(self.function)
#                self.vorticity.data[...] = methodInt(self.function.apply,
#                                                     t, self.dtstretch,
#                                                     self.vorticity.data)
                self.vorticity.data[0][...], self.vorticity.data[1][...], \
                    self.vorticity.data[2][...] = methodInt(
                        self.function.apply,
                        t, self.dtstretch,
                        self.vorticity.data)
                self.synchro.apply()
#            print 'norm omega', LA.norm(self.vorticity.data)

    def stabilityTest(self, dt, maxi, num_method, cststretch):
        dtstretch = dt
        dtstretch = min(dtstretch, cststretch / maxi)
        if dt == dtstretch:
            print "dt, ndt , dtstretch, upperbound", dt, int(dt / dtstretch),\
                  dt / float(int(dt / dtstretch)), cststretch / maxi
            return int(dt / dtstretch)
        else:
            print "dt, ndt , dtstretch, upperbound", dt, int(dt / dtstretch)\
                  + 1, dt / float(int(dt / dtstretch) + 1), cststretch / maxi
            return int(dt / dtstretch) + 1

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Stretching_d"
    print Stretching_d.__doc__
