"""
@file operator/discrete/updateGhosts.py

Update ghost points for some fields defined on a specific topology.
"""
from parmepy.constants import debug, PERIODIC, PARMES_MPI_REAL
from discrete import DiscreteOperator
import numpy as np


class UpdateGhosts_d(DiscreteOperator):
    """
    Ghost points synchronization.
    """

    @debug
    def __init__(self, variables, topo, method=None):
        """
        Constructor.
        Defines the process to update ghosts points of a given list
        of discrete variables. All variables must have the same topology.

        @param  variables : list of discrete fields that must be updated.
        @param topo : the topo on which ghosts are updated.
        @param method : which type of exchange is used.
        """
        DiscreteOperator.__init__(self, variables, method)
        self.input = self.variables
        self.output = self.variables
        self.topology = topo
        for v in self.variables:
            assert v.topology == self.topology, 'different topologies\
                in updateGhost operator.'

        self.ghosts = self.topology.ghosts
        # Indices of points to be filled from previous neighbour
        # Each component is a slice and corresponds to a direction in
        # the topology grid.
        self._g_fromprevious = []
        # Indices of points to be filled from next neighbour.
        self._g_fromnext = []
        # Indices of points to be sent to next neighbour.
        self._g_tonext = []
        # Indices of points to be sent to previous neighbour.
        self._g_toprevious = []
        # Buffers that contains ghosts points to be sent (list, one per dir)
        self._sendbuffer = []
        # Buffers for reception (list, one per dir)
        self._recvbuffer = []
        # domain dimension
        self._dim = self.topology.domain.dimension
        defslice = [slice(None, None, None)] * self._dim

        for d in xrange(self._dim):
            self._g_fromprevious.append(list(defslice))
            self._g_fromprevious[d][d] = slice(self.ghosts[d])
            self._g_fromnext.append(list(defslice))
            self._g_fromnext[d][d] = slice(-self.ghosts[d], None, None)
            self._g_toprevious.append(list(defslice))
            self._g_toprevious[d][d] = slice(self.ghosts[d],
                                             2 * self.ghosts[d], None)
            self._g_tonext.append(list(defslice))
            self._g_tonext[d][d] = slice(-2 * self.ghosts[d], -self.ghosts[d])

            otherDim = [x for x in xrange(self._dim) if x != d]
            for d2 in otherDim:
                self._g_fromprevious[d][d2] = slice(self.ghosts[d2],
                                                    -self.ghosts[d2])
                self._g_fromnext[d][d2] = slice(self.ghosts[d2],
                                                -self.ghosts[d2])
                self._g_toprevious[d][d2] = slice(self.ghosts[d2],
                                                  -self.ghosts[d2])
                self._g_tonext[d][d2] = slice(self.ghosts[d2],
                                              -self.ghosts[d2])

        if self.topology.size > 1:  # else no need to set buffers ...
            # Size computation below assumes that what we send in one
            # dir has the same size as what we receive from process in the
            # same dir ...
            exchange_dir = [d for d in xrange(self._dim)
                            if self.topology.cutdir[d]]
            for d in exchange_dir:
                buffsize = 0
                for v in self.variables:
                    buffsize += v[0][self._g_tonext[d]].size \
                        * v.nbComponents

                self._sendbuffer.append(np.zeros((buffsize)))
                self._recvbuffer.append(np.zeros((buffsize)))

    def applyBC(self, direc=None):
        """
        Setup ghost points for physical boundary conditions
        @param dir : direction where bc must be applied (default = all)
        """
        assert (self.topology.domain.boundaries == PERIODIC).all(),\
            'Only implemented for periodic boundary conditions.'

        if direc is not None:
            dirs = [direc]
        else:
            dirs = xrange(self._dim)
        for v in self.variables:
            for di in xrange(v.nbComponents):
                for d in dirs:
                    v[di][self._g_fromprevious[d]] = \
                        v[di][self._g_tonext[d]]
                    v[di][self._g_fromnext[d]] = v[di][self._g_toprevious[d]]

    @debug
    def setUp(self):
        pass

    @debug
    def apply(self, simulation=None):

        comm = self.topology.comm
        rank = self.topology.rank
        neighbours = self.topology.neighbours
        exchange_dir = [d for d in xrange(self._dim)
                        if self.topology.cutdir[d]]
        i = 0
        for d in exchange_dir:
            # 1 - Fill in buffers
            # Loop through all variables that are distributed
            pos = 0
            nextpos = 0
            for v in self.variables:
                for dir in xrange(v.nbComponents):
                    nextpos += v[dir][self._g_tonext[d]].size
                    self._sendbuffer[i][pos:nextpos] = \
                        v[dir][self._g_tonext[d]].flat
                    pos = nextpos
            # 2 - Send to next receive from previous
            dest_rk = neighbours[1, i]
            from_rk = neighbours[0, i]
            comm.Sendrecv([self._sendbuffer[i], PARMES_MPI_REAL],
                          dest=dest_rk, sendtag=rank,
                          recvbuf=self._recvbuffer[i],
                          source=from_rk, recvtag=from_rk)

            # 3 - Print recvbuffer back to variables and update sendbuffer
            # for next send
            pos = 0
            nextpos = 0
            for v in self.variables:
                for dir in xrange(v.nbComponents):
                    nextpos += v[dir][self._g_fromprevious[d]].size
                    v[dir][self._g_fromprevious[d]].flat = \
                        self._recvbuffer[i][pos:nextpos]
                    self._sendbuffer[i][pos:nextpos] = \
                        v[dir][self._g_toprevious[d]].flat
                    pos = nextpos

            # 4 -Send to previous and receive from next
            dest_rk = neighbours[0, i]
            from_rk = neighbours[1, i]
            comm.Sendrecv([self._sendbuffer[i], PARMES_MPI_REAL],
                          dest=dest_rk, sendtag=rank,
                          recvbuf=self._recvbuffer[i],
                          source=from_rk, recvtag=from_rk)
            # 5 - Print recvbuffer back to variables.
            pos = 0
            nextpos = 0
            for v in self.variables:
                for dir in xrange(v.nbComponents):
                    nextpos += v[dir][self._g_fromprevious[d]].size
                    v[dir][self._g_fromnext[d]].flat = \
                        self._recvbuffer[i][pos:nextpos]
                    pos = nextpos
            # update index in neighbours list
            i += 1
            # End of loop through send/recv directions.

        # Apply boundary conditions for non-distributed directions
        dirs = [d for d in xrange(self._dim)
                if not self.topology.cutdir[d]]

        for d in dirs:
            self.applyBC(d)

if (__name__ == "__main__"):
    print __doc__
    print "- Provided class : UpdateGhosts"
    print UpdateGhosts_d.__doc__
