# -*- coding: utf-8 -*-
"""
@file energy_enstrophy.py
Compute Energy and Enstrophy
"""
import numpy as np
from parmepy.constants import debug, PARMES_MPI_REAL
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.mpi import main_rank, main_size, MPI
from parmepy.tools.timers import timed_function


class Energy_enstrophy(Monitoring):
    """
    Computes enstrophy and the kinetic energy
    \f{eqnarray*}
    enstrophy = \frac{1}{\Omega}\int_\Omega \omega^2 d\Omega
    \f} with \f$\Omega\f$ the volume or surface of the physical domain
    \f$\omega\f$ the vorticity and
    \f{eqnarray*}
    energy = \frac{1}{2\Omega}\int_\Omega v^2 d\Omega
    \f}
    """

    def __init__(self, velocity, vorticity,
                 resolutions=None, viscosity=None,
                 frequency=None, prefix=None):
        """
        Constructor.
        @param fields : List of fields
        @param resolutions :  list of resolutions (one per variable)
        @param viscosity :  kinematic viscosity
        @param frequency : output file producing frequency.
        @param prefix : file name prefix, contains relative path.
        """
        Monitoring.__init__(self, [velocity, vorticity], frequency,
                            name="Energy Enstrophy")
        if prefix is None:
            self.prefix = './Energy'
        else:
            self.prefix = prefix

        self.velocity = velocity
        self.vorticity = vorticity
        self.resolutions = resolutions
        self.viscosity = viscosity
        self.buffer_1 = 0.
        self.buffer_2 = 0.
        if (main_rank == 0):
            self.f = open(self.prefix, 'w')
        self.input = [velocity, vorticity]
        self.output = []

    def setUp(self):
        Monitoring.setUp(self)
        discreteFields = {}
        self.spaceStep = {}
        topodims = np.ones((self.domain.dimension))
        topodims[-1] = main_size

        # Variables discretization
        for v in self.variables:
            # the topology for v ...
            topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                   self.resolutions[v],
                                                   topodims)
            # ... and the corresponding discrete field
            discreteFields[v] = v.discretize(topo)
            self.spaceStep[v] = topo.mesh.space_step
        self.velo = discreteFields[self.velocity]
        self.vort = discreteFields[self.vorticity]

        self._isUpToDate = True

    @debug
    @timed_function
    def apply(self, simulation=None):
        """
        Computation of kinetic energy, enstrophy &
        Checking energy and enstrophy decay
        """
        t = simulation.time
        dt = simulation.timeStep
        ite = simulation.currentIteration
        # Kinetic energy computation
        dvol = np.prod(self.spaceStep[self.velocity])

        if (self.velo.dimension == 2):
            squareNormVel = np.sum(self.velo.data[0] ** 2 +
                                   self.velo.data[1] ** 2) * dvol
        elif (self.velo.dimension == 3):
            squareNormVel = np.sum(self.velo.data[0] ** 2 +
                                   self.velo.data[1] ** 2 +
                                   self.velo.data[2] ** 2) * dvol
        else:
            raise ValueError("invalid problem dimension")

        energy = (0.5 * squareNormVel) / (2.0 * np.pi) ** 3

        # Enstrophy computation
        dvol = np.prod(self.spaceStep[self.vorticity])

        if (self.vort.dimension == 2):
            enstrophy = np.sum(self.vort.data ** 2) * dvol
        elif (self.vort.dimension == 3):
            enstrophy = np.sum(self.vort.data[0] ** 2 +
                               self.vort.data[1] ** 2 +
                               self.vort.data[2] ** 2) * dvol
        else:
            raise ValueError("invalid problem dimension")

        enstrophy = enstrophy / (2.0 * np.pi) ** 3

        # Collective communications

        topovelo = self.velo.topology
        energy = topovelo.comm.reduce(energy, PARMES_MPI_REAL,
                                      op=MPI.SUM, root=0)
        energyBuff1 = topovelo.comm.reduce(self.buffer_1, PARMES_MPI_REAL,
                                           op=MPI.SUM, root=0)
        energyBuff2 = topovelo.comm.reduce(self.buffer_2, PARMES_MPI_REAL,
                                           op=MPI.SUM, root=0)

        topovort = self.vort.topology
        enstrophy = topovort.comm.reduce(enstrophy, PARMES_MPI_REAL,
                                         op=MPI.SUM, root=0)

        # Update buffers

        self.buffer_2 = self.buffer_1
        self.buffer_1 = energy

        # Effective viscosity computation

        if main_rank == 0:

            # 1st order
    #        effectiveViscosity = - (energy - energyBuff1) / (dt * enstrophy)
    #        timeDerivativeEnergy = - (energy -  energyBuff1) / dt

            # 2nd order
            effectiveViscosity = - (3.0 * energy - 4.0 *
                                    energyBuff1 + 1.0 * energyBuff2) / \
                (2.0 * dt * enstrophy)

            ratio = effectiveViscosity / self.viscosity

            # Energy dissipation computation

            timeDerivativeEnergy = - (3.0 * energy - 4.0 * energyBuff1 + 1.0 *
                                      energyBuff2) / (2.0 * dt)

            nuEnstrophy = self.viscosity * enstrophy

            nu_effEnstrophy = effectiveViscosity * enstrophy

    #        dissipE = main_comm.reduce(nuEnstrophy,
    #  PARMES_MPI_REAL, op=MPI.SUM, root=0)

        if ((main_rank == 0) and (ite % self.freq == 0) and (t > dt)):
            self.f = open(self.prefix, 'a')
            self.f.write("%s   %s \
                %s   %s   %s   %s   %s\n" % (t, energy,
                                             enstrophy,
                                             ratio,
                                             timeDerivativeEnergy,
                                             nuEnstrophy,
                                             nu_effEnstrophy))
            self.f.close()

    def __str__(self):
        s = "Energy_enstrophy. "
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Energy_enstrophy"
    print Energy_enstrophy.__doc__
