# -*- coding: utf-8 -*-
"""
@file compute_forces.py
Compute forces
"""
from parmepy.constants import debug, np, PARMES_REAL, PARMES_INTEGER, PI
from parmepy.mpi.topology import Cartesian
from parmepy.numerics.fct2op import Fct2Op
from parmepy.numerics.differentialOperator import DifferentialOperator
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.mpi import main_comm, main_rank, MPI
from parmepy.tools.timers import timed_function


class Compute_forces(Monitoring):
    """
    Compute the forces according the Noca s formula
    """

    def __init__(self, velocity, vorticity,
                 resolutions=None, penal=None,
                 boxMin= None, boxMax=None, Reynolds=None,
                 method='', frequency=None, outputPrefix=None):
        """
        Constructor.
        @param velocity : Continuous velocity field
        @param vorticity : Continuous vorticity field
        @param resolutions :  list of resolutions (one per variable)
        @param penal :  penalization operator
        @param boxMin : Global minimum coordinates of the control volume
        @param boxMax : Global maximum coordinates of the control volume
        @param Reynolds : Reynolds number value
        @param method : numerical method for Differential operators discretizations
        @param frequency : output file producing frequency.
        @param outputPrefix : file name prefix, contains relative path.
        """
        Monitoring.__init__(self, [velocity, vorticity], frequency,
                            name="Forces Computation")
        self.velocity = velocity
        self.vorticity = vorticity
        self.resolutions = resolutions
        self.penal = penal
        self.obstacle = self.penal.obstacle
        self.boxMin = boxMin
        self.boxMax = boxMax
        if not (boxMin < boxMax):
            raise ValueError("Error, boxMin has to be lower than " +
                             "boxMax to define control Noca box")
        self.Re = Reynolds
        self.method = method
        self.outputPrefix = outputPrefix
        self.input = [velocity, vorticity]
        self.output = []

    def setUp(self):
        """
        Transport operator discretization method.
        Create an discrete Transport operator from given specifications.

        """
        Monitoring.setUp(self)
        self.discreteFields = {}

        nbGhosts = 0
        if self.method.find('FD_order2') >= 0:
            nbGhosts = 1
        elif self.method.find('FD_order4') >= 0:
            nbGhosts = 2
        else:
            nbGhosts = 2

        # Variables discretization
        ghosts = None
        for v in self.variables:
            if (v==self.velocity):
                ghosts = np.zeros((self.domain.dimension))
                ghosts[:] = nbGhosts
            # the topology for v ...
            topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                   self.resolutions[v],
                                                   ghosts=ghosts)

            # ... and the corresponding discrete field
            self.discreteFields[v] = v.discretize(topo)
        self.velo = self.discreteFields[self.velocity]
        self.vort = self.discreteFields[self.vorticity]

#        # velocity discretization
#        topoId, topo = self.velocity.domain.addTopology(
#            Cartesian(domain=self.domain, dim=self.penal.domain.dimension,
#                      globalMeshResolution=self.penal.resolutions[self.velocity],
#                      ghosts=[nbGhosts, nbGhosts, nbGhosts]))
#        df, dfId = self.velocity.discretize(topo)
#        self.penal.discreteFieldId[self.velocity] = dfId
#        self.velocity = self.velocity.discreteField[
#            self.penal.discreteFieldId[self.penal.velocity]]
#        print 'topo', self.velocity.topology

#        # vorticity discretization
#        topoId, topo = self.vorticity.domain.addTopology(
#            Cartesian(domain=self.domain, dim=self.penal.domain.dimension,
#                      globalMeshResolution=self.penal.resolutions[self.vorticity]))
#        df, dfId = self.vorticity.discretize(topo)
#        self.penal.discreteFieldId[self.vorticity] = dfId
#        self.vorticity = self.vorticity.discreteField[
#            self.penal.discreteFieldId[self.penal.vorticity]]

        self.compute_control_box()
        self._isUpToDate = True

        if (main_rank == 0):
            self.f = open(self.outputPrefix, 'w')

    def compute_control_box(self) :
        """
        Compute indicator functions for the control box
        (including the obstacle)

        The computation of index is based on velocity topology
        """
        self.dim = self.velo.topology.dim
        self.res = self.velo.topology.mesh.resolution
        self.step = self.velo.topology.mesh.space_step
        self.ghosts = self.velo.topology.ghosts
        self.periods = self.velo.topology.periods
        self.coord_start = self.velo.topology.mesh.origin + \
                        (self.ghosts * self.step)
        self.coord_end = self.velo.topology.mesh.end - \
                        (self.ghosts * self.step)

#        print 'res, dx, cMin, cMax, bmin, bMax', self.res, self.step, self.coord_start, self.coord_end, self.boxMin, self.boxMax

        # normal vectors definitions
        self.normal = np.zeros([self.dim*2, self.dim], dtype=PARMES_REAL)
        self.Up = 0
        self.Down = 1
        self.East = 2
        self.West = 3
        self.North = 4
        self.South = 5
        self.normal[self.Up][0] = 1.0
        self.normal[self.Down][0] = - 1.0
        self.normal[self.East][1] = 1.0
        self.normal[self.West][1] = - 1.0
        self.normal[self.North][2] = 1.0
        self.normal[self.South][2] = - 1.0

        # control box definition
        tmp_ind = np.zeros([self.dim, 2], dtype=PARMES_INTEGER)

        distMin = self.boxMin - self.coord_start
        distMax = self.boxMax - self.coord_end
        print 'coord_start, coord_end', self.coord_start, self.coord_end
        print 'distMin, distMax, step, res', distMin, distMax, self.step, self.res
        for i in xrange(self.dim):
            if (distMin[i]>=0. and distMax[i]<=0.):
            # the control volume is included inside the local domain
                tmp_ind[i][0] = distMin[i] // self.step[i] # Down, West, South
                tmp_ind[i][1] = self.res[i] -2. * self.ghosts[i] + distMax[i] // self.step[i] -1# approximation : lower integer
            if (distMin[i]>=0. and self.boxMin[i]<= self.coord_end[i] and distMax[i]>=0.):
            # only min corner of control volume is included inside the local domain
                tmp_ind[i][0] = distMin[i] // self.step[i]
                tmp_ind[i][1] = self.res[i] -2. * self.ghosts[i] - 1
            if (distMin[i]<=0. and self.boxMax[i]>= self.coord_start[i] and distMax[i]<=0.) :
            # only max corner of control volume is included inside the local domain
                tmp_ind[i][1] = self.res[i] -2.*self.ghosts[i] + distMax[i] // self.step[i] -1
            if (distMin[i]<=0. and distMax[i]>=0.):
            # the local domain is included inside the control volume
                tmp_ind[i][1] = self.res[i] - 2. * self.ghosts[i] - 1


        ind = np.zeros(self.dim * 2, dtype=PARMES_INTEGER)
        ind[self.Down] = tmp_ind[0][0]
        ind[self.Up] = tmp_ind[0][1]
        ind[self.West] = tmp_ind[1][0]
        ind[self.East] = tmp_ind[1][1]
        ind[self.South] = tmp_ind[2][0]
        ind[self.North] = tmp_ind[2][1]
        print 'ind_Down, ind_Up', ind[self.Down], ind[self.Up]
        print 'ind_West, ind_East', ind[self.West], ind[self.East]
        print 'ind_South, ind_North', ind[self.South], ind[self.North]

        if ind.all() == 0:
            # there is no volume control inside the local domain --> pass
            self.nbPointsBox = 0

        else:
            # Count the number of points on each face and inside the domain
            nbPoints = np.zeros(2 * self.dim, dtype=PARMES_INTEGER)
            nbPoints[self.Down] = (ind[self.East] - ind[self.West] + 1) * \
                                  (ind[self.North] - ind[self.South] + 1)
            nbPoints[self.Up] = (ind[self.East] - ind[self.West] + 1) * \
                                (ind[self.North] - ind[self.South] + 1)
            nbPoints[self.West] = (ind[self.Up] - ind[self.Down] + 1) * \
                                  (ind[self.North] - ind[self.South] + 1)
            nbPoints[self.East] = (ind[self.Up] - ind[self.Down] + 1) * \
                                  (ind[self.North] - ind[self.South] + 1)
            nbPoints[self.South] = (ind[self.East] - ind[self.West] + 1) * \
                                   (ind[self.Up] - ind[self.Down] + 1)
            nbPoints[self.North] = (ind[self.East] - ind[self.West] + 1) * \
                                   (ind[self.Up] - ind[self.Down] + 1)

            print 'nb points plan Down', nbPoints[self.Down]

            boxDim = np.zeros(self.dim, dtype=PARMES_INTEGER)
            boxDim[0] = ind[self.Up] - ind[self.Down] + 1
            boxDim[1] = ind[self.East] - ind[self.West] + 1
            boxDim[2] = ind[self.North] - ind[self.South] + 1
            self.nbPointsBox = boxDim[0] * boxDim[1] * boxDim[2]

            tmp_box = np.zeros([self.nbPointsBox,self.dim], dtype=PARMES_INTEGER)
#            chi_box_i=[]
#            chi_box_j=[]
#            chi_box_k=[]
            coords = np.zeros(self.dim, dtype=PARMES_REAL)
            count_box = 0

            for k in xrange (ind[self.South],ind[self.North] + 1): ## delete loops !!
                coords[2] = self.coord_start[2] + k * self.step[2]
                for j in xrange (ind[self.West],ind[self.East] + 1):
                    coords[1] = self.coord_start[1] + j * self.step[1]
                    for i in xrange (ind[self.Down],ind[self.Up] + 1):
                        coords[0] = self.coord_start[0] + i * self.step[0]
                        dist = np.vdot(coords - self.obstacle.center,
                                       coords - self.obstacle.center) \
                               - self.obstacle.definedObst.radius ** 2
                        if(dist >= 0.0 - 1.E-12): # We are on or outside the sphere ...
                           tmp_box[count_box] = [i, j, k]
                           count_box = count_box + 1
#                            chi_box_i.append(i)
#                            chi_box_j.append(j)
#                            chi_box_k.append(k)

#            print 'tmp_box', tmp_box, tmp_box.shape, count_box

            # local indices of "COMPUTED" points in the control box which are outside the obstacle
            self.chi_box = np.zeros([count_box,self.dim], dtype=PARMES_INTEGER)
            self.chi_box = tmp_box[0:count_box,:] ### PBM !!!

#            chi_box_i = np.asarray(chi_box_i, dtype=PARMES_INTEGER)
#            chi_box_j = np.asarray(chi_box_j, dtype=PARMES_INTEGER)
#            chi_box_k = np.asarray(chi_box_k, dtype=PARMES_INTEGER)
#            self.chi_box = tuple([chi_box_i, chi_box_j, chi_box_k])

            self.chi_up = np.zeros([nbPoints[self.Up],self.dim], dtype=PARMES_INTEGER)
            self.chi_down = np.zeros([nbPoints[self.Down],self.dim], dtype=PARMES_INTEGER)
            self.chi_east = np.zeros([nbPoints[self.East],self.dim], dtype=PARMES_INTEGER)
            self.chi_west = np.zeros([nbPoints[self.West],self.dim], dtype=PARMES_INTEGER)
            self.chi_north = np.zeros([nbPoints[self.North],self.dim], dtype=PARMES_INTEGER)
            self.chi_south = np.zeros([nbPoints[self.South],self.dim], dtype=PARMES_INTEGER)

            # local indices of "COMPUTED" points located on the surfaces of the control box

            if(nbPoints[self.South]!=0.): # South boundary is in the domain
                count = 0
                for j in xrange (ind[self.West],ind[self.East] + 1):
                    for i in xrange (ind[self.Down],ind[self.Up] + 1):
                        self.chi_south[count] = [i, j, ind[self.South]]
                        count = count + 1

            if(nbPoints[self.East]!=0.): # East boundary is in the domain
                count = 0
                for k in xrange (ind[self.South],ind[self.North] + 1):
                    for i in xrange (ind[self.Down],ind[self.Up] + 1):
                        self.chi_east[count] = [i, ind[self.East], k]
                        count = count + 1

            if(nbPoints[self.Down]!=0.): # Down boundary is in the domain
                count = 0
                for k in xrange (ind[self.South],ind[self.North] + 1):
                    for j in xrange (ind[self.West],ind[self.East] + 1):
                        self.chi_down[count] = [ind[self.Down], j, k]
                        count = count + 1

            if(nbPoints[self.North]!=0.): # North boundary is in the domain
                count = 0
                for j in xrange (ind[self.West],ind[self.East] + 1):
                    for i in xrange (ind[self.Down],ind[self.Up] + 1):
                        self.chi_north[count] = [i, j, ind[self.North]]
                        count = count + 1

            if(nbPoints[self.West]!=0.): # West boundary is in the domain
                count = 0
                for k in xrange (ind[self.South],ind[self.North] + 1):
                    for i in xrange (ind[self.Down],ind[self.Up] + 1):
                        self.chi_west[count] = [i, ind[self.West], k]
                        count = count + 1

            if(nbPoints[self.Up]!=0.): # Up boundary is in the domain
                count = 0
                for k in xrange (ind[self.South],ind[self.North] + 1):
                    for j in xrange (ind[self.West],ind[self.East] + 1):
                        self.chi_up[count] = [ind[self.Up], j, k]
                        count = count + 1

#            # South boundary is in the domain
#            chi_south_i=[]
#            chi_south_j=[]
#            chi_south_k=[]
#            if(nbPoints[self.South]!=0.):
#                for j in xrange (ind[self.West],ind[self.East] + 1):
#                    for i in xrange (ind[self.Down],ind[self.Up] + 1):
#                        chi_south_i.append(i)
#                        chi_south_j.append(j)
#                        chi_south_k.append(ind[self.South])
#            chi_south_i = np.asarray(chi_south_i, dtype=PARMES_INTEGER)
#            chi_south_j = np.asarray(chi_south_j, dtype=PARMES_INTEGER)
#            chi_south_k = np.asarray(chi_south_k, dtype=PARMES_INTEGER)
#            self.chi_south = tuple([chi_south_i, chi_south_j, chi_south_k])

#            # East boundary is in the domain
#            chi_east_i=[]
#            chi_east_j=[]
#            chi_east_k=[]
#            if(nbPoints[self.East]!=0.):
#                for k in xrange (ind[self.South],ind[self.North] + 1):
#                    for i in xrange (ind[self.Down],ind[self.Up] + 1):
#                        chi_east_i.append(i)
#                        chi_east_j.append(ind[self.East])
#                        chi_east_k.append(k)
#            chi_east_i = np.asarray(chi_east_i, dtype=PARMES_INTEGER)
#            chi_east_j = np.asarray(chi_east_j, dtype=PARMES_INTEGER)
#            chi_east_k = np.asarray(chi_east_k, dtype=PARMES_INTEGER)
#            self.chi_east = tuple([chi_east_i, chi_east_j, chi_east_k])

#            # Down boundary is in the domain
#            chi_down_i=[]
#            chi_down_j=[]
#            chi_down_k=[]
#            if(nbPoints[self.Down]!=0.):
#                for k in xrange (ind[self.South],ind[self.North] + 1):
#                    for j in xrange (ind[self.West],ind[self.East] + 1):
#                        chi_down_i.append(ind[self.Down])
#                        chi_down_j.append(j)
#                        chi_down_k.append(k)
#            chi_down_i = np.asarray(chi_down_i, dtype=PARMES_INTEGER)
#            chi_down_j = np.asarray(chi_down_j, dtype=PARMES_INTEGER)
#            chi_down_k = np.asarray(chi_down_k, dtype=PARMES_INTEGER)
#            self.chi_down = tuple([chi_down_i, chi_down_j, chi_down_k])

#            # North boundary is in the domain
#            chi_north_i=[]
#            chi_north_j=[]
#            chi_north_k=[]
#            if(nbPoints[self.North]!=0.):
#                for j in xrange (ind[self.West],ind[self.East] + 1):
#                    for i in xrange (ind[self.Down],ind[self.Up] + 1):
#                        chi_north_i.append(i)
#                        chi_north_j.append(j)
#                        chi_north_k.append(ind[self.North])
#            chi_north_i = np.asarray(chi_north_i, dtype=PARMES_INTEGER)
#            chi_north_j = np.asarray(chi_north_j, dtype=PARMES_INTEGER)
#            chi_north_k = np.asarray(chi_north_k, dtype=PARMES_INTEGER)
#            self.chi_north = tuple([chi_north_i, chi_north_j, chi_north_k])

#            # West boundary is in the domain
#            chi_west_i=[]
#            chi_west_j=[]
#            chi_west_k=[]
#            if(nbPoints[self.West]!=0.):
#                for k in xrange (ind[self.South],ind[self.North] + 1):
#                    for i in xrange (ind[self.Down],ind[self.Up] + 1):
#                        chi_west_i.append(i)
#                        chi_west_j.append(ind[self.West])
#                        chi_west_k.append(k)
#            chi_west_i = np.asarray(chi_west_i, dtype=PARMES_INTEGER)
#            chi_west_j = np.asarray(chi_west_j, dtype=PARMES_INTEGER)
#            chi_west_k = np.asarray(chi_west_k, dtype=PARMES_INTEGER)
#            self.chi_west = tuple([chi_west_i, chi_west_j, chi_west_k])

#            # Up boundary is in the domain
#            chi_up_i=[]
#            chi_up_j=[]
#            chi_up_k=[]
#            if(nbPoints[self.Up]!=0.):
#                for k in xrange (ind[self.South],ind[self.North] + 1):
#                    for j in xrange (ind[self.West],ind[self.East] + 1):
#                        chi_up_i.append(ind[self.Up])
#                        chi_up_j.append(j)
#                        chi_up_k.append(k)
#            chi_up_i = np.asarray(chi_up_i, dtype=PARMES_INTEGER)
#            chi_up_j = np.asarray(chi_up_j, dtype=PARMES_INTEGER)
#            chi_up_k = np.asarray(chi_up_k, dtype=PARMES_INTEGER)
#            self.chi_up = tuple([chi_up_i, chi_up_j, chi_up_k])

        print 'nb points box', self.nbPointsBox

        self.bufferForce = 0.
        # Compute coef used to determine drag coefficient : cD = coef.Fx = 2.Fx/rho.uinf**2.D
        uinf = 1.
        self.coef = 2. / (uinf ** 2 * PI * self.obstacle.definedObst.radius ** 2)

    @debug
    @timed_function
    def apply(self, simulation):
        """
        Computation of the drag according to the "impulsion" formula presented by
        Noca et. al in Journal of Fluids and Structures, 1999
        """
        t = simulation.time
        dt = simulation.timeStep
        ite = simulation.currentIteration

        if self.nbPointsBox == 0:
            pass
        else :
            localForce = np.zeros(self.dim, dtype=PARMES_REAL)
            force = np.zeros(self.dim, dtype=PARMES_REAL)
            # computation of hessian matrices (d2_u1, d2_u2 and d2_u3) and jacobian matrix (d_u)
            hessian1, hessian2, hessian3 = Fct2Op(self.velo, self.velo, self.method,
                                                  choice = 'hessianU').apply(None,None)
            self.jacobian, maxgersh = DifferentialOperator(self.velo, self.velo, self.method,
                                                           choice='gradV').__call__()
            # computation of the components of Div(T) where T = 1/Re * (Nabla(U) + Nabla(U)T)
            self.a = 1. / self.Re * (2. * hessian1[0] + hessian1[4] + hessian2[1] + \
                                     hessian1[8] + hessian3[2])
            self.b = 1. / self.Re * (hessian2[0] + hessian1[1] + 2. * hessian2[4] + \
                                     hessian2[8] + hessian3[5])
            self.c = 1. / self.Re * (hessian3[0] + hessian1[6] + hessian3[4] + \
                                     hessian2[7] + 2. * hessian3[8])

            # Downstream and upstream surfaces
            dsurf = self.step[1] * self.step[2]
            localForce = self.integrateOnSurface(localForce, self.velo, self.vort, self.chi_down,
                                                 self.normal[self.Down][:], self.Re, dsurf)
            localForce = self.integrateOnSurface(localForce, self.velo, self.vort, self.chi_up,
                                                 self.normal[self.Up][:], self.Re, dsurf)
            # East and West surfaces
            dsurf = self.step[0] * self.step[2]
            localForce = self.integrateOnSurface(localForce, self.velo, self.vort, self.chi_east,
                                                 self.normal[self.East][:], self.Re, dsurf)
            localForce = self.integrateOnSurface(localForce, self.velo, self.vort, self.chi_west,
                                                 self.normal[self.West][:], self.Re, dsurf)
            # North and south surfaces
            dsurf = self.step[0] * self.step[1]
            localForce = self.integrateOnSurface(localForce, self.velo, self.vort, self.chi_south,
                                                 self.normal[self.South][:], self.Re, dsurf)
            localForce = self.integrateOnSurface(localForce, self.velo, self.vort, self.chi_north,
                                                 self.normal[self.North][:], self.Re, dsurf)
            # over the volume
            dvol = self.step[0] * self.step[1] * self.step[2]
            localForce = self.integrateOnBox(localForce, self.vort, self.chi_box, dvol, dt)

            localForce = localForce * self.coef
            comm = main_comm
            comm.Reduce([localForce, MPI.DOUBLE], [force, MPI.DOUBLE], op=MPI.SUM, root=0)

            # print time and forces values in the following order : time, cX, cY, cZ
            if ((main_rank == 0) and (ite % self.freq == 0) and (t > dt)):
                self.f = open(self.outputPrefix, 'a')
                self.f.write("%s   %s   %s   %s\n" % (t, force[0], force[1], force[2]))
                self.f.close()

    def integrateOnBox(self, force, vort, chi, dvol, dt):
    ## WARNING : this routine is available inly if velo and vort
    ## topologies have the same global mesh resolution !!!
        """
        Return -1/(dim-1)*d/dt int_over_control_box coord X vorticity
        """
        fact = - dvol / ((self.dim - 1) * dt)
        int1 = np.zeros(self.dim, dtype=PARMES_REAL)
        # For all points in the box
        for ind in xrange (chi.shape[1]):
            l = chi[ind][0]
            m = chi[ind][1]
            n = chi[ind][2]
            # coordinates of the current point
            coords = self.coord_start + chi[ind] * self.step
            # part1 of the force
            int1 = int1 + np.vdot(coords,vort[l,m,n])

        force = force + fact * (int1 - self.bufferForce) # 1st order time integration
        self.bufferForce = int1 # Save for next time step ...

        return force

    def integrateOnSurface(self, force, velo, vort, chi, NormalVec, Re, dsurf):
    ## WARNING : this routine is available inly if velo and vort
    ## topologies have the same global mesh resolution !!!
        """
        Compute integrals on surface to calculate forces acting on the body.
        See (2.1) of Noca 1999 or (52) of Plouhmans 2002
        Integrals on the obstacle are neglected.
        """
        fact = 1. / (self.dim - 1)
        int1 = np.zeros(self.dim, dtype=PARMES_REAL)
        int2 = np.zeros(self.dim, dtype=PARMES_REAL)
        X_n_DivT = np.zeros(self.dim, dtype=PARMES_REAL)
        n_T = np.zeros(self.dim, dtype=PARMES_REAL)
        # For each point of the current plane
        for ind in xrange (chi.shape[0]):
            l = chi[ind][0]
            m = chi[ind][1]
            n = chi[ind][2]
            # adjust indices for velocity vector field (which contains ghosts points)
            i = l + self.ghosts[0]
            j = m + self.ghosts[1]
            k = n + self.ghosts[2]

            # part1 = 1/2(velocity.velocity)n - (n.velocity)velocity - 1/(dim-1)((n.velocity)(coord X vorticity) + 1/(dim-1)(n.vorticity)(coord X velocity)

            # (velocity.velocity)
            u_u = np.vdot(velo[l,m,n],velo[l,m,n])
            # normal.velocity
            n_u = np.vdot(velo[l,m,n],NormalVec[:])
            # normal.vorticity
            n_w = np.vdot(vort[i,j,k],NormalVec[:])
            # coordinates of the current point
            coords = self.coord_start[:] + chi[ind] * self.step[:]
            # part1 of the force
            int1 = int1 + 0.5 * u_u * NormalVec[:] - n_u * velo[l,m,n] \
                - fact * n_u * np.cross(coords,vort[i,j,k]) \
                + fact * n_w * np.cross(coords,velo[l,m,n])


            # part 2 of the force : part 2 = 1/(dim-1)(X^(n^Div(T)) + n.T
            if (chi.all() == self.chi_down.all()): # normalVec = (-1,0,0)
                X_n_DivT[0] = - coords[1] * self.b[i,j,k] - coords[2] * self.c[i,j,k]
                X_n_DivT[1] = coords[0] * self.b[i,j,k]
                X_n_DivT[2] = coords[0] * self.c[i,j,k]
                n_T[0] = - 1. / Re * 2 * self.jacobian[0][i,j,k]
                n_T[1] = - 1. / Re * (self.jacobian[3][i,j,k] + self.jacobian[1][i,j,k])
                n_T[2] = - 1. / Re * (self.jacobian[6][i,j,k] + self.jacobian[2][i,j,k])

            if (chi.all() == self.chi_up.all()): # normalVec = (1,0,0)
                X_n_DivT[0] = coords[1] * self.b[i,j,k] + coords[2] * self.c[i,j,k]
                X_n_DivT[1] = - coords[0] * self.b[i,j,k]
                X_n_DivT[2] = - coords[0] * self.c[i,j,k]
                n_T[0] = 1. / Re * 2 * self.jacobian[0][i,j,k]
                n_T[1] = 1. / Re * (self.jacobian[3][i,j,k] + self.jacobian[1][i,j,k])
                n_T[2] = 1. / Re * (self.jacobian[6][i,j,k] + self.jacobian[2][i,j,k])

            if (chi.all() == self.chi_west.all()): # normalVec = (0,-1,0)
                X_n_DivT[0] = coords[1] * self.a[i,j,k]
                X_n_DivT[1] = - coords[0] * self.a[i,j,k] - coords[2] * self.c[i,j,k]
                X_n_DivT[2] = coords[1] * self.c[i,j,k]
                n_T[0] = - 1. / Re * (self.jacobian[1][i,j,k] + self.jacobian[3][i,j,k])
                n_T[1] = - 1. / Re * 2 * self.jacobian[4][i,j,k]
                n_T[2] = - 1. / Re * (self.jacobian[7][i,j,k] + self.jacobian[5][i,j,k])

            if (chi.all() == self.chi_east.all()): # normalVect = (0,1,0)
                X_n_DivT[0] = - coords[1] * self.a[i,j,k]
                X_n_DivT[1] = coords[0] * self.a[i,j,k] + coords[2] * self.c[i,j,k]
                X_n_DivT[2] = - coords[1] * self.c[i,j,k]
                n_T[0] = 1. / Re * (self.jacobian[1][i,j,k] + self.jacobian[3][i,j,k])
                n_T[1] = 1. / Re * 2 * self.jacobian[4][i,j,k]
                n_T[2] = 1. / Re * (self.jacobian[7][i,j,k] + self.jacobian[5][i,j,k])

            if (chi.all() == self.chi_south.all()): # normalVec = (0,0,-1)
                X_n_DivT[0] = coords[2] * self.a[i,j,k]
                X_n_DivT[1] = coords[2] * self.b[i,j,k]
                X_n_DivT[2] = - coords[0] * self.a[i,j,k] - coords[1] * self.b[i,j,k]
                n_T[0] = - 1. / Re * (self.jacobian[2][i,j,k] + self.jacobian[6][i,j,k])
                n_T[1] = - 1. / Re * (self.jacobian[5][i,j,k] + self.jacobian[7][i,j,k])
                n_T[2] = - 1. / Re * 2 * self.jacobian[8][i,j,k]

            if (chi.all() == self.chi_north.all()): # normalVec = (0,0,1)
                X_n_DivT[0] = - coords[2] * self.a[i,j,k]
                X_n_DivT[1] = - coords[2] * self.b[i,j,k]
                X_n_DivT[2] = coords[0] * self.a[i,j,k] + coords[1] * self.b[i,j,k]
                n_T[0] = 1. / Re * (self.jacobian[2][i,j,k] + self.jacobian[6][i,j,k])
                n_T[1] = 1. / Re * (self.jacobian[5][i,j,k] + self.jacobian[7][i,j,k])
                n_T[2] = 1. / Re * 2 * self.jacobian[8][i,j,k]

            int2 = int2 + fact * X_n_DivT + n_T
        # Product with element of surface and sum to the total (input) force
        result = force + (int1 + int2) * dsurf

        return result

    def __str__(self):
        s = "Compute_forces. "
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Compute_forces"
    print Compute_forces.__doc__
