# -*- coding: utf-8 -*-
"""
@file operator/stretching.py

Stretching operator representation


"""
from parmepy.constants import debug
from parmepy.operator.continuous import Operator
from parmepy.operator.discrete.stretching import Stretching_d
import numpy as np


class Stretching(Operator):
    """
    Stretching operator representation
    \todo write latex formulas
    """

    @debug
    def __init__(self, velocity, vorticity,
                 resolutions=None, method='FD_order4 RK3',
                 topo=None, ghosts=None, **other_config):
        """
        Constructor.
        Create a Stretching operator from given
        velocity and vorticity variables.

        @param velocity ContinuousVectorField : velocity variable.
        @param vorticity ContinuousVectorField : vorticity variable.
        """
        Operator.__init__(self, [velocity, vorticity], method, topo=topo,
                          ghosts=ghosts, name="Stretching")
        ## velocity variable (vector)
        self.velocity = velocity
        ## vorticity variable (vector)
        self.vorticity = vorticity
        self.resolutions = resolutions
        self.config = other_config
        self.input = [self.velocity, self.vorticity]
        self.output = [self.vorticity]

    @debug
    def setUp(self):
        """
        Stretching operator discretization method.
        Create a discrete Stretching operator from given specifications.

        @param idVelocityD : Index of velocity discretisation to use.
        @param idVorticityD : Index of vorticity discretisation to use.
        @param propertyOp : property garantied by the stretching Op
            (divConservation, massConservation) default : divConservation
        @param method : the method to use :
            space : (FD_order2, FD_order4, ..) default : FD_order4
            and time : (Euler, RK, ..) default : RK3
        """
        print '================ Stretching scheme ====================='
        print ' Space discretization and integr. scheme:', self.method
        print '========================================================'
        nbGhosts = 0
        if self.method.find('FD_order2') >= 0:
            nbGhosts = 1
        elif self.method.find('FD_order4') >= 0:
            nbGhosts = 2
        else:
            nbGhosts = 2

        if self._predefinedTopo is not None:
            assert (self._predefinedTopo.ghosts >= nbGhosts).all()
            topo = self._predefinedTopo
            for v in self.variables:
                self.discreteFields[v] = v.discretize(topo)
        else:
            # Variables discretization
            if self.ghosts is not None:
                self.ghosts[self.ghosts < nbGhosts] = nbGhosts
            else:
                self.ghosts = np.zeros((self.domain.dimension)) * nbGhosts
            for v in self.variables:
                # the topology for v ...
                topo = self.domain.getOrCreateTopology(self.domain.dimension,
                                                       self.resolutions[v],
                                                       ghosts=self.ghosts)

                # ... and the corresponding discrete field
                self.discreteFields[v] = v.discretize(topo)

        self.discreteOperator = \
            Stretching_d(self.discreteFields[self.velocity],
                         self.discreteFields[self.vorticity],
                         method=self.method, **self.config)

        self.discreteOperator.setUp()
        self._isUpToDate = True

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Stretching"
    print Stretching.__doc__
