"""
@file simulation.py

Description of the simulation parameters (time, iteration ...)
"""
from parmepy.mpi import main_rank


class Simulation(object):
    """
    Setup for simulation parameters.
    """

    def __init__(self, tinit=0.0, tend=1.0, timeStep=0.01, iterMax=100):
        """
        Creates a Timer.

        @param tend : Simulation final time.
        @param timeStep : Time step.
        @param tinit : Simulation starting time.
        @param iterMax : Simulation maximum iteration number.
        """
        ## Simulation final time
        self.end = tend
        ## Starting time
        self.start = tinit
        ## Simulation current time
        self.time = tinit
        ## Is simulation is terminated
        self.isOver = False
        ## Iteration counter
        self.currentIteration = 1
        ## Simulation time step
        self.timeStep = timeStep
        ## Maximum number of iterations
        self.iterMax = iterMax

    def advance(self):
        """
        Proceed to next time
        """
        if self.currentIteration < self.iterMax:
            self.currentIteration += 1
            self.time += self.timeStep
            # adjust last timestep to reach self.end
            if self.end - self.time < self.timeStep and self.end > self.time:
                self.timeStep = self.end - self.time
        else:
            self.isOver = True
        if self.time >= self.end:
            self.isOver = True

    def printState(self):
        """
        print current state
        """
        if main_rank == 0:
            print "=== Iteration : {0:3d}, start from t ={1:6.3f} \
to t ={2:6.3f} ===".format(
                self.currentIteration, self.time, self.time + self.timeStep)

    def reset(self):
        """
        set initial time to current and reset iteration counter.
        Used to initialize simulation after a restart.
        """
        self.start = self.time
        self.isOver = False
        self.currentIteration = 1

    def __str__(self):
        s = "Simulation parameters : "
        s += "from " + str(self.start) + ' to ' + str(self.end)
        s += ', time step : ' + str(self.timeStep)
        s += ', current time : ' + str(self.time) + ', iteration number : '
        s += str(self.currentIteration) + ', max number of iterations : '
        s += str(self.iterMax)
        return s
