"""
@file timers.py

Contains class for monitoring computational time in a non-intrusive way
thanks to decorator.
"""
from parmepy.mpi import MPI, main_rank
from parmepy import __VERBOSE__


def timed_function(f):
    """
    Decorator for function timing. Get the corresponding function timer in the
    operator (represented in args[0]).

    @remark : Works only with methods that belongs to object with Timer
    member named timer.
    """
    def wrapper(*args, **kwargs):
        f_timer = args[0].timer.getFunctionTimer(f)
        res = f_timer(f)(*args, **kwargs)
        return res
    return wrapper


class FunctionTimer(object):
    """
    Class for timing functions with the timed_function decorator.

    The FunctionTimer is instancied with a given function. Calling the timer
    will call the function timed with execution time computation.
    """

    def __init__(self, func):
        """
        Creates a timer for the given function.

        @param func : function to time.
        """
        ## Timed function name
        self.func_name = func.func_name
        ## Total time spent in the function (in seconds).
        self.t = 0.
        ## Times per calls
        self.times = []

    def __call__(self, func):
        """
        The function given is wrapped with time computations instructions
        through MPI.Wtime function.
        It returns the wrapped function.
        @param func function to time.
        @return wrapped function.
        """
        def f(*args, **kwargs):
            t0 = MPI.Wtime()
            res = func(*args, **kwargs)
            t = MPI.Wtime() - t0
            self.t += t
            self.times.append(t)
            if __VERBOSE__:
                print args[0].name, ' -- ', self.func_name, " : ", t, 's'
            return res
        return f

    def __str__(self):
        s = "[" + str(main_rank) + "]                |-- " + self.func_name
        s += ' = ' + str(self.t) + " s  ({0} calls)".format(len(self.times))
        return s


class ManualFunctionTimer(FunctionTimer):
    """
    Class for manual timing some code. For instance, this class is desined for
    monitoring OpenCL kernels where function call does not reflect
    computational time.
    """
    def __init__(self, name):
        def fun():
            pass
        fun.func_name = name
        FunctionTimer.__init__(self, fun)

    def append_time(self, t):
        """Manual computational time append"""
        self.t += t
        self.times.append(t)
        if __VERBOSE__:
            print self.func_name, " : ", t, 's'


class Timer(object):
    """
    Manages a dictionary of FunctionTimer objects for monitoring functions
    member of the given operator.
    """

    def __init__(self, op):
        """
        Creates an Timer.
        @param op : Timer owner.
        """
        ## Parent object name
        self._obj = op
        ## FunctionTimer dictionary with Functions names as keys.
        self.f_timers = {}
        ## Total time spent in the operator (in seconds).
        self.t = 0.

    def __add__(self, t):
        """
        Overrides operator +.
        @param t : Other Timer object
        """
        t._compute_summary()
        self.t += t.t
        for t_k in t.f_timers.keys():
            t.f_timers[t_k].func_name = \
                t._obj.name + "_" + t.f_timers[t_k].func_name
            self.f_timers[t._obj.name + "_" + t_k] = t.f_timers[t_k]
        return self

    def getFunctionTimer(self, func):
        """
        Get the FunctionTimer related to given function, or created
        if not exists.

        @param func : Function to get the corresponding timer.
        @return : FunctionTimer related to func.
        """
        try:
            return self.f_timers[func.func_name]
        except KeyError:
            self.f_timers[func.func_name] = FunctionTimer(func)
            return self.f_timers[func.func_name]

    def addFunctionTimer(self, ft):
        """
        Add a function timer to Timer.

        @param ft : FunctionTimer to add.
        """
        self.f_timers[ft.func_name] = ft

    def _compute_summary(self):
        """
        Copute a summary of the different functions referenced herein.
        """
        for f in self.f_timers.keys():
            self.t += self.f_timers[f].t

    def __str__(self):
        self._compute_summary()
        s = "[" + str(main_rank) + "] Time spent in "
        s += self._obj.name + ' = ' + str(self.t) + ' s'
        for f in sorted(self.f_timers.keys()):
            s += "\n" + str(self.f_timers[f])
        return s

    def toString(self):
        """
        Returns a string containing timings informations
        @return timings as string.
        """
        h = self._obj.name.replace(' ', '_') + ' '
        t = str(self.t) + ' '
        l = sorted(self.f_timers.keys())
        h += ' '.join(l) + ' '
        t += ' '.join([str(self.f_timers[k].t) for k in l]) + ' '
        return h, t
